"""
The challenge here is how to make classes that can quickly ingress data in a format convenient
to the rollout worker, store the data efficiently,
and equally quickly egress the data in a format useful to the various algorithms.
Ideally we would have a minimal set of storage classes that can provide good performance across all uses.
"""
import collections
import datetime
import pathlib
import pickle
import random
import shutil
import uuid
from collections import defaultdict
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import numba
import numpy as np
import torch
from loguru import logger
from torch import Tensor
from tree import map_structure

from avalon.agent.common.types import BatchSequenceData
from avalon.agent.common.types import SequenceData
from avalon.agent.common.types import StepData
from avalon.agent.common.util import create_action_storage
from avalon.agent.common.util import create_observation_storage
from avalon.agent.dreamer.params import OffPolicyParams
from avalon.agent.dreamer.params import Params
from avalon.datagen.world_creation.constants import int_to_avalon_task

try:
    import pyarrow
except ImportError:
    pass


class ModelStorage:
    # TODO: move this somewhere better + parameterize
    # TODO: make this have a unique path so we can run multiple runs on the same machine
    model_path = Path("/tmp/model.pt")

    @classmethod
    def clean(cls) -> None:
        cls.model_path.unlink(missing_ok=True)

    @classmethod
    def push_model(cls, model: torch.nn.Module) -> None:
        """Save the model to storage."""
        torch.save(model.state_dict(), str(cls.model_path))

    @classmethod
    def load_model(cls, model: torch.nn.Module, device: torch.device) -> None:
        """Load the parameters in storage into the provided model."""
        if cls.model_path.exists():
            try:
                model.load_state_dict(torch.load(str(cls.model_path), map_location=device))
                logger.info("loaded model")
            except Exception as e:
                # Most commonly we fail here if the file is being written while we're reading.
                # RuntimeError: PytorchStreamReader failed reading zip archive: failed finding central directory
                # EOFError: Ran out of input
                logger.warning(f"exception loading model in worker: {e}")
        else:
            logger.info("no model found")


class TrajectoryStorage:
    def __init__(self, params: Params, step_data_type: Type[StepData], num_workers: int) -> None:
        self.params = params
        self.num_workers = num_workers
        self.step_data_type = step_data_type

    def partial_batch_update(
        self,
        key: str,
        value: Union[dict, Tensor],
        mask: Tensor,
        timesteps: Tensor,
        is_extra_observation: bool = False,
    ) -> None:
        """This ingress method is optimized for the current rollout worker.
        The worker always passes a full batch of values, but not all are valid, nor are they all on the same timestep.
        So we pass extra fields giving this necessary info to construct a valid update to storage.
        - `key`: the storage key, eg `observation` or `done`
        - `value`: a batch of values (one for each rollout worker) (or a dict of those if the key has nested values)
        - `mask`: a boolean array indicating which values should be written to storage
        - `timesteps`: an integer array, indicating which timestep into the episode or fragment the value corresponds to.
        - `is_extra_observation`: on-policy algs require n+1 observations, for boostrapping; this flag indicates if this
            is that extra final observation. That obs should be ignored on other storage types.
        """
        raise NotImplementedError

    def mark_step_finished(self, done_idxs) -> None:
        """Indicate that a given rollout step is over; ie we have received a full set of keys.
        The expectation is that between calling this method, each key will be updated exactly once."""
        raise NotImplementedError

    def reset(self) -> None:
        pass

    def shutdown(self) -> None:
        pass


class EpisodeStorage(TrajectoryStorage):
    def __init__(  # type: ignore[no-untyped-def]
        self,
        params: OffPolicyParams,
        step_data_type: Type[StepData],
        num_workers: int,
        rollout_dir: Optional[str] = None,
        wandb_queue=None,  # not typeable, but a Optional[Queue] :(
        episode_callback: Optional[Callable[[list[StepData]], None]] = None,
        discard_short_eps: bool = True,  # should be true in training, false in eval
        in_memory_buffer_size: Optional[int] = None,  # if none, set to params.log_rollout_metrics_every
    ) -> None:
        """Episode-based storage.
        - if `rollout_dir` is None, won't save episodes to disk
        - if `wandb_queue` is None, won't log stats
        - if a `episode_callback` is given, will run this hook on every completed episode
        """
        super().__init__(params, step_data_type, num_workers)
        self.wandb_queue = wandb_queue
        self.episode_callback = episode_callback
        self.discard_short_eps = discard_short_eps

        self.store_to_disk = rollout_dir is not None
        if self.store_to_disk:
            self.data_dir = Path(rollout_dir)
            self.data_dir.mkdir(parents=True, exist_ok=True)
            logger.info(self.data_dir)

        if in_memory_buffer_size is None:
            in_memory_buffer_size = params.log_rollout_metrics_every
        self.recent_eps: collections.deque[List[StepData]] = collections.deque(maxlen=in_memory_buffer_size)

        # {worker_id: {key: value, ...}, ...}
        self.current_step_data = defaultdict(lambda: defaultdict(dict))
        self.ongoing_episodes: list[List[StepData]] = [[] for _ in range(num_workers)]

        self.timesteps_in_buffer = 0
        self.total_timesteps = 0
        self.episode_counter = 0
        self.last_logged_episode = 0

    def partial_batch_update(
        self,
        key: str,
        value: Union[dict, Tensor],
        mask: Tensor,
        timesteps: Tensor,
        is_extra_observation: bool = False,
    ) -> None:
        if is_extra_observation:
            return

        # This code is a little weird - where we might expect "scalars", are actually single-element tensors,
        # because slicing one item from a 1-d tensor gives a single-item tensor (unlike in numpy).
        # But it is helpful for one reason, which is that the dtype is preserved -
        # a fp32 stays a fp32 vs if it were converted to a python float and then back to tensor,
        # it would end up as a fp64 tensor by default.

        # copying/cloning the tensor is necessary, since otherwise we're just saving a reference
        # to a specific slice of a tensor that is (potentially) reused in the worker.

        # Annoyingly there's not a great way to use a `map_structure` command here since the target atoms
        # can be immutable, or nonexisting. Thus the duplication.
        if isinstance(value, Tensor):
            for i, v in enumerate(value):
                if mask[i]:
                    self.current_step_data[i][key] = v.clone()
        elif isinstance(value, dict):
            for k2, v2 in value.items():
                for i, v in enumerate(v2):
                    if mask[i]:
                        self.current_step_data[i][key][k2] = v.clone()

    def mark_step_finished(self, done_idxs) -> None:
        """The worker uses this to indicate that it's finished with a rollout step."""
        for worker_id in [worker_id for worker_id, done in enumerate(done_idxs) if done]:
            v = self.current_step_data[worker_id]
            del self.current_step_data[worker_id]
            # StepData takes an info, don't need to filter it out here
            v = self.step_data_type(**v)
            self.ongoing_episodes[worker_id].append(v)
            if v.done:
                ep = self.ongoing_episodes[worker_id]
                self.ongoing_episodes[worker_id] = []
                ep_len = len(ep)
                if self.discard_short_eps and len(ep) < self.params.min_fragment_len:
                    logger.debug(f"not adding short ep of len {ep_len}")
                    continue
                self.episode_counter += 1
                # We do these before packing the ep because the infos are lost in the current packing code.
                self.recent_eps.append(ep)
                if self.episode_callback:
                    self.episode_callback(ep)
                packed_ep = v.pack_sequence(ep)

                self.timesteps_in_buffer += ep_len
                self.total_timesteps += ep_len
                self.save_episode(packed_ep)
                self.log_stats()

    def log_stats(self) -> None:
        # Compute training rollout statistics
        if not self.wandb_queue:
            return
        if not self.params.log_rollout_metrics_every:
            return
        if (self.episode_counter - self.last_logged_episode) < self.params.log_rollout_metrics_every:
            return
        self.last_logged_episode = self.episode_counter

        if self.params.env_params.suite == "godot":
            # Can't use log_rollout_stats because these are full episodes, not fragments,
            # so the analysis is a bit different. We can make stronger statements here.
            successes: dict[str, dict[str, Any]] = defaultdict(lambda: defaultdict(list))
            keys = ["success", "difficulty"]
            for episode in self.recent_eps:
                last_step = episode[-1]
                task = int_to_avalon_task[int(last_step.info["task"])].lower()
                for field in keys:
                    successes[task][f"final_{field}"].append(last_step.info[field])
                successes[task][f"episode_length"].append(len(episode))
            # Data is a dict (task) of dicts (keys) of lists
            for task, x in successes.items():
                for field, y in x.items():
                    self.wandb_queue.put(("scalar", f"{task}/{field}", np.mean(y)))
                self.wandb_queue.put(("scalar", f"{task}/num_episodes", len(y)))
        else:
            rewards = []
            lengths = []
            for episode in self.recent_eps:
                total_reward = sum([step.reward for step in episode])
                length = len(episode)
                rewards.append(total_reward)
                lengths.append(length)
            # Note: action repeats have not been factored in here
            self.wandb_queue.put(("scalar", f"return_mean", np.mean(rewards)))
            self.wandb_queue.put(("scalar", f"length_mean", np.mean(lengths)))

        self.wandb_queue.put(("scalar", f"buffer_size_timesteps", self.timesteps_in_buffer))
        self.wandb_queue.put(("scalar", f"total_timesteps", self.total_timesteps))
        if self.store_to_disk:
            self.wandb_queue.put(("scalar", f"true_buffer_size", measure_buffer_size(str(self.data_dir))))
        self.wandb_queue.put(("scalar", f"total_episodes", self.episode_counter))

    def save_episode(self, episode: SequenceData) -> None:
        """Write an episode to disk."""
        if self.store_to_disk:
            # if self.params.observation_compression:
            #     for k in ("rgb", "rgbd"):
            #         if k in episode.observation:
            #             episode.observation[k] = pack(episode.observation[k])
            timestamp = datetime.datetime.now().strftime("%Y%m%dT%H%M%S")
            identifier = str(uuid.uuid4().hex)
            length = len(episode.done)
            filename = self.data_dir / f"{timestamp}-{identifier}-{length}.npz"
            with filename.open("wb") as f2:
                pickle.dump(episode, f2)

            self.enforce_buffer_size()

    def enforce_buffer_size(self) -> None:
        # enforce timestep limit
        # TODO: kinda weird that we will be deleting other processes' files
        # TODO: if one of these classes crashes, then the allowed buffer size will grow
        # probably should just count all the filenames instead - was just trying to avoid that.
        # because nobody is keeping an absolute track of the size of the buffer.
        if self.timesteps_in_buffer > self.params.replay_buffer_size_timesteps_per_manager * 1.1:
            files = [p.name for p in pathlib.Path(self.data_dir).iterdir() if p.is_file()]
            random.shuffle(files)
            while self.timesteps_in_buffer > self.params.replay_buffer_size_timesteps_per_manager:
                file = files.pop()
                try:
                    (self.data_dir / file).unlink()
                    self.timesteps_in_buffer -= parse_filename(str(file))["timesteps"]
                except FileNotFoundError:
                    # This happens if two processes are cleaning up simultaneously.
                    continue

    def shutdown(self) -> None:
        if self.store_to_disk:
            shutil.rmtree(str(self.data_dir), ignore_errors=True)


@numba.njit(parallel=False, cache=True)
def numba_assign(target, source, mask, timesteps) -> None:
    """There's not a good way (that I could find) to make this operation fast in torch/numpy natively.
    The operation, in numpy notation, is `target[mask][timesteps] = source[mask]`
    So this numba function provides a pretty significant speedup.
    """
    n = len(target)
    for worker_id in numba.prange(n):
        if mask[worker_id]:
            t = timesteps[worker_id]
            # This actual memory copy is maybe slower in numba than numpy for large arrays?
            target[worker_id, t] = source[worker_id]


class FragmentStorage(TrajectoryStorage):
    """In-memory storage for on-policy / PPO-like algorithms.
    These receive an array of fixed shape (num_envs, num_timesteps) each iteration for training.
    The storage is typically reset between iterations.

    The one exception is that the observation has shape (num_envs, num_timesteps+1), for value boostrapping.
    """

    def __init__(self, params: Params, step_data_type: Type[StepData], num_workers: int) -> None:
        super().__init__(params, step_data_type, num_workers)
        self.storage = self.build_storage()

    def mark_step_finished(self, *args, **kwargs) -> None:
        pass

    def partial_batch_update(
        self, key: str, value: Union[dict, Tensor], mask: Tensor, timesteps: Tensor, is_extra_observation: bool = False
    ) -> None:
        # This codepath is pretty carefully optimized; modify only if you know what you're doing!
        def assign(target, source) -> None:
            args = (target.numpy(), source.numpy(), mask.numpy(), timesteps.numpy())
            numba_assign(*args)

        map_structure(assign, self.storage[key], value)

    def to_packed(self) -> BatchSequenceData:
        return self.step_data_type.batch_sequence_type(**{k: v for k, v in self.storage.items() if k != "info"})

    def reset(self, fast_reset: bool = False, retain_final_obs: bool = True) -> None:
        # This relies on us replacing, not modifying the old observation when resetting.
        # Could do a deepcopy but unnecessary as-is.
        old_observation = self.storage["observation"]

        # A "fast reset" is actually just not resetting anything, lol. Just let the new overwrite the old.
        # Does potetially make debugging more confusing.
        if not fast_reset:
            self.storage = self.build_storage()

        if retain_final_obs:
            # We init obs[t=0] as the final obs from the last batch (that +1 extra obs)
            # This should typically always be used.
            for k, v in old_observation.items():
                self.storage["observation"][k][:, 0] = v[:, -1]

    def build_storage(self):
        storage = {
            # Note the extra timestep here
            "observation": create_observation_storage(
                self.params.observation_space, batch_shape=(self.num_workers, self.params.num_steps + 1)
            ),
            "action": create_action_storage(
                self.params.action_space, batch_shape=(self.num_workers, self.params.num_steps)
            ),
            "reward": torch.zeros((self.num_workers, self.params.num_steps), dtype=torch.float32),
            "value": torch.zeros((self.num_workers, self.params.num_steps), dtype=torch.float32),
            "policy_prob": torch.zeros((self.num_workers, self.params.num_steps), dtype=torch.float32),
            "policy_entropy": torch.zeros((self.num_workers, self.params.num_steps), dtype=torch.float32),
            "done": torch.zeros((self.num_workers, self.params.num_steps), dtype=torch.bool),
            "is_terminal": torch.zeros((self.num_workers, self.params.num_steps), dtype=torch.bool),
            "info": {
                k: torch.zeros((self.num_workers, self.params.num_steps), dtype=torch.float32)
                for k in self.params.env_params.info_fields
            },
        }
        return storage


def pack(data: Any):  # type: ignore
    """Compress any python type."""
    import lz4.frame

    # Observed compression ratio on synthetic godot numpy data: 17MB -> 2-3MB
    data = pyarrow.serialize(data).to_buffer().to_pybytes()
    data = lz4.frame.compress(data)
    return data


def unpack(data):  # type: ignore
    """Uncompress something compressed with `pack`."""
    import lz4.frame

    data = lz4.frame.decompress(data)
    data = pyarrow.deserialize(data)
    return data


def parse_filename(filename: str) -> Dict:
    """Split a filename into its components."""
    # 20220610T123952-edd6a7dfbff544939976663265fb4147-40.npz
    timestamp, hash, timesteps = filename.split(".")[0].split("-")
    return {"timestamp": timestamp, "hash": hash, "timesteps": int(timesteps)}


def measure_buffer_size(folder: str) -> int:
    """Calculate the total number of frames in the given folder."""
    folder_path = Path(folder)
    files = [p.name for p in pathlib.Path(folder_path).iterdir() if p.is_file()]
    return sum([parse_filename(filename)["timesteps"] for filename in files])



from typing import Optional

import attr
import torch
from loguru import logger

from avalon.agent.common.models import ActivationFunction
from avalon.agent.common.params import ProcgenEnvironmentParams
from avalon.agent.common.parse_args import parse_args
from avalon.agent.common.trainer import OnPolicyTrainer
from avalon.agent.common.util import setup_new_process
from avalon.agent.ppo.params import ObservationModelParams
from avalon.agent.ppo.params import PPOParams
from avalon.common.log_utils import configure_remote_logger

NUM_WORKERS = 64
NUM_STEPS = 256


@attr.s(auto_attribs=True, frozen=True)
class ProcgenPPOParams(PPOParams):
    project: str = "zack_zack_ppo"
    total_env_steps: int = 25_000_000
    num_steps: int = NUM_STEPS
    num_workers: int = NUM_WORKERS
    batch_size: int = int(NUM_STEPS * NUM_WORKERS / 8)
    ppo_epochs: int = 3
    discount: float = 0.999
    lam: float = 0.95
    value_loss_coef: float = 0.5
    entropy_coef: float = 0.01
    clip_range: float = 0.2
    clip_range_vf: Optional[float] = 0.2
    baselines_style_vf_loss: bool = True
    lr: float = 5e-4
    clip_grad_norm: float = 0.5
    time_limit_bootstrapping: bool = False
    env_params: ProcgenEnvironmentParams = ProcgenEnvironmentParams()
    log_freq_hist: int = 500
    log_freq_scalar: int = 50
    log_freq_media: int = 500
    checkpoint_every: int = 2500
    model_params: ObservationModelParams = ObservationModelParams(
        mlp_activation_fn=ActivationFunction.ACTIVATION_TANH,  # not used
        num_mlp_layers=1,  # no MLP, just adapting shape from 256 to output dim
        encoder_output_dim=256,
        num_cnn_base_channels=16,
    )

    eval_exploration_mode: str = "eval"  # TODO: try explore here also
    center_observations: bool = False
    center_and_clamp_discrete_logits: bool = False


def run(params: ProcgenPPOParams) -> None:
    trainer = OnPolicyTrainer(params)
    try:
        if params.is_training:
            trainer.train()
            trainer.shutdown()
            if trainer.train_storage:
                trainer.train_storage.reset()
            torch.cuda.empty_cache()  # just for seeing what's going on
        if params.is_testing:
            pass
    finally:
        trainer.shutdown()


if __name__ == "__main__":
    configure_remote_logger(level="INFO")
    setup_new_process()
    try:
        default_params = ProcgenPPOParams()
        default_params = parse_args(default_params)
        run(default_params)
    except Exception as e:
        logger.exception(e)
        raise



import math
import os
import struct
import time
from collections import OrderedDict
from io import BufferedReader
from io import BufferedWriter
from math import prod
from threading import Thread
from types import TracebackType
from typing import Any
from typing import Dict
from typing import Generic
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Type
from typing import cast

import numpy as np
from loguru import logger
from numpy import typing as npt

from avalon.datagen.errors import GodotError
from avalon.datagen.godot_env.actions import ActionType
from avalon.datagen.godot_env.actions import DebugCameraAction
from avalon.datagen.godot_env.actions import _to_bytes
from avalon.datagen.godot_env.interactive_godot_process import InteractiveGodotProcess
from avalon.datagen.godot_env.observations import FAKE_TYPE_IMAGE
from avalon.datagen.godot_env.observations import NP_DTYPE_MAP
from avalon.datagen.godot_env.observations import FeatureSpecDict
from avalon.datagen.godot_generated_types import ACTION_MESSAGE
from avalon.datagen.godot_generated_types import BRIDGE_LOG_SIGNAL
from avalon.datagen.godot_generated_types import CLOSE_MESSAGE
from avalon.datagen.godot_generated_types import DEBUG_CAMERA_ACTION_MESSAGE
from avalon.datagen.godot_generated_types import LOAD_SNAPSHOT_MESSAGE
from avalon.datagen.godot_generated_types import QUERY_AVAILABLE_FEATURES_MESSAGE
from avalon.datagen.godot_generated_types import RENDER_MESSAGE
from avalon.datagen.godot_generated_types import RESET_MESSAGE
from avalon.datagen.godot_generated_types import SAVE_SNAPSHOT_MESSAGE
from avalon.datagen.godot_generated_types import SEED_MESSAGE
from avalon.datagen.godot_generated_types import SELECT_FEATURES_MESSAGE

# Data dictionary of features, interpretable by a `FeatureSpecDict`
FeatureDataDict = Dict[str, npt.NDArray]

_BINARY_READ: Literal["br"] = "br"
_BINARY_WRITE: Literal["bw"] = "bw"


class _BridgeKillSwitch:
    """Kills the GodotProcess if a blocking action fails to complete.

    Starts a background process that will kill the godot_process if an any
    logic wrapped in a `with watch_blocking_action()` fails to complete within
    the timeout.
    """

    def __init__(
        self,
        godot_process: InteractiveGodotProcess,
        check_period_seconds: float = 5.0,
        default_timeout: float = 30.0,
    ) -> None:
        self._kill_time: float = math.inf
        self._godot_process = godot_process
        self._default_timeout = default_timeout
        self._kill_error: Optional[GodotError] = None

        self._check_period_seconds = check_period_seconds
        self._thread = Thread(
            target=self._kill_if_blocked,
            daemon=True,
        )
        self._thread.start()

    def watch_blocking_action(self, timeout_seconds: Optional[float] = None):
        self._kill_time = time.time() + (timeout_seconds or self._default_timeout)
        return self

    def __enter__(self) -> "_BridgeKillSwitch":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self._kill_time = math.inf
        if self._kill_error is None:
            return
        if exc_val is not None:
            exc_val.__cause__ = self._kill_error
            return
        raise self._kill_error

    def raise_any_logged_errors(self) -> None:
        self._godot_process.raise_any_logged_godot_errors()

    def _kill_if_blocked(self) -> None:
        while True:
            time.sleep(self._check_period_seconds)
            if self._godot_process.is_finished:
                return
            if self._godot_process.is_running and time.time() > self._kill_time:
                logger.error(f"_BridgeKillSwitch killing godot: blocking action took too long")
                try:
                    self._godot_process.check_for_errors()
                    self._kill_error = GodotError(f"_BridgeKillSwitch killing process")
                    self._godot_process.close(kill=True)
                except GodotError as e:
                    self._kill_error = GodotError(f"_BridgeKillSwitch killed process: {e.args[0]}")
                return


class GodotEnvBridge(Generic[ActionType]):
    """Low-level byte bridge to Godot leveraging unix pipes."""

    def __init__(
        self,
        action_pipe: BufferedWriter,
        action_record_pipe: BufferedWriter,
        observation_pipe: BufferedReader,
        kill_switch: _BridgeKillSwitch,
        close_timeout_seconds: float,
        screen_shape: Tuple[int, int, int],
    ) -> None:
        self._action_pipe = action_pipe
        self._action_record_pipe = action_record_pipe
        self._observation_pipe = observation_pipe
        self._kill_switch = kill_switch
        self._close_timeout_seconds = close_timeout_seconds

        self._screen_shape = screen_shape
        self._selected_features: FeatureSpecDict = OrderedDict()

    @classmethod
    def _kill_on_parsing_errors_to_avoid_zombies(cls, process: InteractiveGodotProcess):
        try:
            process.wait_for_log_signal(BRIDGE_LOG_SIGNAL)
        except (GodotError, TimeoutError):
            process.close(kill=True)
            raise

    @classmethod
    def build_by_starting_process(
        cls,
        process: InteractiveGodotProcess,
        screen_resolution: Tuple[int, int],
        build_timeout_seconds: float = 30.0,
        close_timeout_seconds: float = 30.0,
    ) -> "GodotEnvBridge":
        """Create pipes while starting process so that the bridge doesn't block.

        Waits for godot to send a ready signal before returning
        """
        kill_switch = _BridgeKillSwitch(process)
        with kill_switch.watch_blocking_action(build_timeout_seconds):
            _mkpipe(process.action_pipe_path)
            _mkpipe(process.observation_pipe_path)

            process.start()

            cls._kill_on_parsing_errors_to_avoid_zombies(process)

            bridge = cls(
                action_pipe=open(process.action_pipe_path, _BINARY_WRITE),
                action_record_pipe=open(process.action_record_path, _BINARY_WRITE),
                observation_pipe=open(process.observation_pipe_path, _BINARY_READ),
                kill_switch=kill_switch,
                close_timeout_seconds=close_timeout_seconds,
                screen_shape=(*screen_resolution, 3),
            )

            process.wait_for_ready_signal()

        return bridge

    @property
    def is_open(self) -> bool:
        return not (self._action_pipe.closed or self._observation_pipe.closed or self._action_record_pipe.closed)

    def query_available_features(self) -> FeatureSpecDict:
        with self._kill_switch.watch_blocking_action():
            self._send_message(QUERY_AVAILABLE_FEATURES_MESSAGE, bytes())

            feature_count = self._read_int()
            available_features: FeatureSpecDict = OrderedDict()
            for _ in range(feature_count):
                feature_name_len = self._read_int()
                feature_name = self._observation_pipe.read(feature_name_len).decode("UTF-8")
                data_type = self._read_int()
                dim_count = self._read_int()
                dims = tuple(self._read_int() for x in range(dim_count))
                available_features[feature_name] = (data_type, dims)
            return available_features

    def select_and_cache_features(self, selected_features: FeatureSpecDict) -> None:
        self._selected_features = selected_features
        with self._kill_switch.watch_blocking_action():
            size_doubleword = (len(selected_features)).to_bytes(4, byteorder="little", signed=False)
            feature_names_bytes = ("\n".join(selected_features.keys()) + "\n").encode("UTF-8")
            self._send_message(SELECT_FEATURES_MESSAGE, size_doubleword + feature_names_bytes)

    def seed(self, episode_seed: int) -> None:
        with self._kill_switch.watch_blocking_action():
            message_bytes = episode_seed.to_bytes(8, byteorder="little", signed=True)
            self._send_message(SEED_MESSAGE, message_bytes)

    def reset(
        self, null_action: ActionType, episode_seed: int, world_path: str, starting_hit_points: float
    ) -> FeatureDataDict:
        with self._kill_switch.watch_blocking_action():
            self._send_message(
                RESET_MESSAGE,
                null_action.to_bytes()
                + episode_seed.to_bytes(8, byteorder="little", signed=True)
                + (world_path + "\n").encode("UTF-8")
                + _to_bytes(float, starting_hit_points),
            )
            return self._read_features(self._selected_features)

    def act(self, action: ActionType) -> FeatureDataDict:
        with self._kill_switch.watch_blocking_action():
            self._send_message(ACTION_MESSAGE, action.to_bytes())
            return self._read_features(self._selected_features)

    def debug_act(self, action: DebugCameraAction) -> FeatureDataDict:
        with self._kill_switch.watch_blocking_action():
            self._send_message(DEBUG_CAMERA_ACTION_MESSAGE, action.to_bytes())
            return self._read_features(self._selected_features)

    def render(self) -> npt.NDArray:
        with self._kill_switch.watch_blocking_action():
            self._send_message(RENDER_MESSAGE, bytes())
            return self._read_shape(self._screen_shape, prod(self._screen_shape), dtype=np.uint8)

    def save_snapshot(self) -> str:
        with self._kill_switch.watch_blocking_action():
            self._send_message(SAVE_SNAPSHOT_MESSAGE, bytes())
            snapshot_path = self._observation_pipe.readline().decode("UTF-8")[:-1]
            return snapshot_path

    def load_snapshot(self, snapshot_path: str) -> FeatureDataDict:
        with self._kill_switch.watch_blocking_action():
            self._send_message(
                LOAD_SNAPSHOT_MESSAGE,
                (snapshot_path + "\n").encode("UTF-8"),
            )
            return self._read_features(self._selected_features)

    def close(self) -> None:
        with self._kill_switch.watch_blocking_action(self._close_timeout_seconds):
            self._send_message(CLOSE_MESSAGE, bytes())

    def after_close(self) -> None:
        try:
            self._action_pipe.close()
        except BrokenPipeError:
            pass
        try:
            self._observation_pipe.close()
        except BrokenPipeError:
            pass
        try:
            self._action_record_pipe.close()
        except BrokenPipeError:
            pass

    def _send_message(self, message_type: int, message: bytes) -> None:
        """Send a message down the godot action pipe.

        This is always a single byte message_type, followed by the message.
        """
        assert isinstance(message, bytes), f"action {cast(Any, message)} must be bytes"
        self._action_pipe_write((message_type).to_bytes(1, byteorder="little", signed=False))
        self._action_pipe_write(message)

        self._action_pipe.flush()
        self._action_record_pipe.flush()

    def _action_pipe_write(self, data: bytes) -> None:
        self._action_pipe.write(data)
        self._action_record_pipe.write(data)

    def _read_int(self) -> int:
        return cast(int, struct.unpack("i", self._observation_pipe.read(4))[0])

    def _read_shape(
        self,
        shape: Tuple[int, ...],
        size: Optional[int] = None,
        dtype: npt.DTypeLike = np.uint8,
        feature_name: Optional[str] = None,
    ) -> npt.NDArray:
        read_size = size if size is not None else prod(shape)
        byte_buffer = self._observation_pipe.read(read_size)
        try:
            return np.ndarray(shape=shape, dtype=dtype, buffer=byte_buffer)
        except TypeError as te:
            self._kill_switch.raise_any_logged_errors()
            raise GodotError(
                f"Invalid number of bytes for feature {feature_name}: "
                f"got {len(byte_buffer)}, expected {read_size} of {dtype}"
            ) from te

    def _read_features(self, selected_features: FeatureSpecDict) -> FeatureDataDict:
        feature_data: FeatureDataDict = {}
        for feature_name, (data_type, dims) in selected_features.items():
            size = prod(dims)
            if data_type != FAKE_TYPE_IMAGE:
                size = size * 4
            feature_data[feature_name] = self._read_shape(
                dims, size, dtype=NP_DTYPE_MAP[data_type], feature_name=feature_name
            )
        return feature_data


def _mkpipe(pipe: str) -> None:
    if os.path.exists(pipe):
        os.remove(pipe)
    os.mkfifo(pipe)



from typing import Optional

import attr

from avalon.contrib.serialization import Serializable
from avalon.datagen.world_creation.types import DebugVisualizationConfig


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ExportConfig(Serializable):
    name: str
    is_tiled: bool
    is_terrain_shader_enabled: bool
    is_flora_shader_enabled: bool
    is_sun_enabled: bool
    is_sun_shadow_enabled: bool
    is_ssao_enabled: bool
    is_fog_enabled: bool
    is_blur_enabled: bool
    is_postprocessing_enabled: bool
    is_indoor_lighting_enabled: bool
    scenery_mode: str
    is_minor_scenery_hidden: bool
    is_biome_fast: bool
    is_exported_with_absolute_paths: bool
    is_border_calculation_detailed: bool
    is_meta_data_exported: bool
    world_id: Optional[str] = None
    debug_visualization_config: Optional[DebugVisualizationConfig] = None
    # TODO(mjr) added back for regression tests. Should probably disentangle them from world generation
    is_legacy_biome_based_on_natural_height_set: Optional[bool] = None


def get_oculus_export_config(world_id: Optional[str] = None) -> ExportConfig:
    return ExportConfig(
        name="oculus",
        is_tiled=True,
        is_terrain_shader_enabled=False,
        is_flora_shader_enabled=False,
        is_sun_enabled=True,
        is_sun_shadow_enabled=True,
        is_ssao_enabled=True,
        is_fog_enabled=True,
        is_blur_enabled=False,
        is_postprocessing_enabled=True,
        is_indoor_lighting_enabled=True,
        scenery_mode="normal",
        is_minor_scenery_hidden=False,
        is_biome_fast=False,
        is_exported_with_absolute_paths=False,
        is_border_calculation_detailed=True,
        is_meta_data_exported=True,
        world_id=world_id,
    )


def get_eval_agent_export_config() -> ExportConfig:
    return attr.evolve(
        get_oculus_export_config(),
        # evaluation is very similar to the oculus config, with only minor (basically graphical) differences
        name="agent",
        # the biggest difference is that the levels are never "tiled" for agents.
        # the tiling takes MUCH longer to export for larger levels, and is likely only marginally faster
        # at runtime, and thus isn't really worth the trade-off (unless you really want to scale your world generation
        # workers out like crazy)
        is_tiled=False,
        # we disable ssao, fog, postprocessing, and lighting so that the simulator is faster
        # none of thes options makes a meaningful difference graphically, but they look nicer for players
        # also, postprocessing is used for players to indicate health
        is_ssao_enabled=False,
        is_fog_enabled=False,
        is_postprocessing_enabled=False,
        is_indoor_lighting_enabled=False,
        # minor scenery (ie, not trees, ie, things that are not collidable) should not be visible to the agent
        # becauase it does not train with them (because they don't make a big difference, and takes much longer
        # to generate all of them). For the evaluation mode, we generate them then hide them so that the level gen
        # code will work out to produce exactly the same levels as with the oculus human export config
        # (while in training we dont even bother generating them at all in the first place)
        is_minor_scenery_hidden=True,
        # the way godot loads levels is different for the agent and the human players, and requires absolute paths
        is_exported_with_absolute_paths=True,
    )


def get_agent_export_config() -> ExportConfig:
    return attr.evolve(
        get_eval_agent_export_config(),
        # training only differs in a very small number of ways
        # first, we just avoid generating any of the non-tree (non-collision) scenery), since the agents don't even
        # see that anyway in the evaluation settings (it's just to make the worlds look nicer for humans)
        scenery_mode="tree",
        # next, we dont care about making super accurate border calculations. They take a while, but dont really
        # change the world that much. Again, just makes it look nicer for players basically.
        is_border_calculation_detailed=False,
        # finally, no need to export the meta data file during training, because we dont use it at all
        is_meta_data_exported=False,
    )


def get_pretty_export_config() -> ExportConfig:
    return ExportConfig(
        name="pretty",
        is_tiled=False,
        is_terrain_shader_enabled=True,
        is_flora_shader_enabled=True,
        is_sun_enabled=True,
        is_sun_shadow_enabled=True,
        is_ssao_enabled=True,
        is_fog_enabled=True,
        is_blur_enabled=True,
        is_postprocessing_enabled=True,
        is_indoor_lighting_enabled=True,
        scenery_mode="normal",
        is_minor_scenery_hidden=False,
        is_biome_fast=False,
        is_exported_with_absolute_paths=False,
        is_border_calculation_detailed=True,
        is_meta_data_exported=True,
    )


def get_mouse_and_keyboard_export_config() -> ExportConfig:
    return ExportConfig(
        name="mouse_and_keyboard",
        is_tiled=False,
        is_terrain_shader_enabled=True,
        is_flora_shader_enabled=True,
        is_sun_enabled=True,
        is_sun_shadow_enabled=True,
        is_ssao_enabled=True,
        is_fog_enabled=True,
        is_blur_enabled=True,
        is_postprocessing_enabled=True,
        is_indoor_lighting_enabled=False,
        scenery_mode="normal",
        is_minor_scenery_hidden=False,
        is_biome_fast=False,
        is_exported_with_absolute_paths=False,
        is_border_calculation_detailed=True,
        is_meta_data_exported=True,
    )



from typing import cast

import attr
import numpy as np

from avalon.common.utils import to_immutable_array
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.entities.food import Food
from avalon.datagen.world_creation.entities.pillar import Pillar
from avalon.datagen.world_creation.entities.utils import get_random_ground_points
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import Point3DListNP
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.worlds.height_map import HeightMap


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Altar(Pillar):
    position: np.ndarray = attr.ib(
        default=np.array([0.0, 0.0, 0.0]),
        converter=to_immutable_array,
        eq=attr.cmp_using(eq=np.array_equal),  # type: ignore[attr-defined]
    )
    entity_id: int = -1
    is_tool_helpful: bool = False

    @staticmethod
    def build(height: float, table_dim: float = 1.0) -> "Altar":
        return Altar(size=np.array([table_dim, height, table_dim]))

    def place(self, food_ground_position: Point3DNP) -> "Altar":
        return attr.evolve(self, position=food_ground_position.copy() + UP_VECTOR * self.get_offset())

    def get_food_height(self, food: Food) -> float:
        return cast(float, self.size[1] + food.get_offset())

    def get_offset(self) -> float:
        return cast(float, self.size[1] / 2.0)

    def get_food_locations(
        self,
        rand: np.random.Generator,
        center: Point3DNP,
        count: int,
        map: HeightMap,
        min_radius: float,
        max_radius: float,
        offset: float,
        island_mask: MapBoolNP,
    ) -> Point3DListNP:
        points = get_random_ground_points(rand, center, count, map, min_radius, max_radius, offset, island_mask)
        points[:, 1] = self.size[1] + offset
        return points



from typing import List

import attr
import numpy as np
from godot_parser import Node as GDNode
from godot_parser import NodePath
from scipy.spatial.transform import Rotation

from avalon.common.utils import only
from avalon.datagen.world_creation.entities.doors.locks.door_lock import DoorLock
from avalon.datagen.world_creation.entities.doors.types import HingeSide
from avalon.datagen.world_creation.geometry import Axis
from avalon.datagen.world_creation.indoor.blocks import new_make_block_nodes
from avalon.datagen.world_creation.indoor.constants import DEFAULT_STORY_HEIGHT
from avalon.datagen.world_creation.indoor.utils import get_scaled_mesh_node
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.utils import make_transform


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class RotatingBar(DoorLock):
    size: np.ndarray = np.array([1, DEFAULT_STORY_HEIGHT, 0.1])
    rotation: np.ndarray = np.array([1, 0, 0, 0, 1, 0, 0, 0, 1])
    rotation_axis: Axis = Axis.Z
    anchor_side: HingeSide = HingeSide.RIGHT
    unlatch_angle: float = 10.0  # degrees

    def get_node(self, scene: GodotScene) -> GDNode:
        ROTATING_BAR = "rotating_bar"
        ROTATING_BAR_BODY = "bar_body"
        ROTATING_BAR_ANCHOR = "anchor"

        anchor_collision_mesh_node = GDNode(
            "collision_shape",
            type="CollisionShape",
            properties={
                "shape": scene.add_sub_resource("SphereShape", radius=0).reference,
                "disabled": True,
            },
        )
        anchor_node = GDNode(ROTATING_BAR_ANCHOR, "StaticBody")
        anchor_node.add_child(anchor_collision_mesh_node)

        joint_lower_angle = 0 if self.anchor_side == HingeSide.RIGHT else -180
        joint_upper_angle = -180 if self.anchor_side == HingeSide.RIGHT else 0
        hinge_joint_node = GDNode(
            "hinge_joint",
            "HingeJoint",
            properties={
                "nodes/node_a": NodePath(f"../{ROTATING_BAR_ANCHOR}"),
                "nodes/node_b": NodePath(f"../{ROTATING_BAR_BODY}"),
                "angular_limit/enable": True,
                "angular_limit/lower": joint_lower_angle,
                "angular_limit/upper": joint_upper_angle,
                "angular_limit/relaxation": 0.25,
            },
        )

        bar_offset = np.array([self.size[0] / 2, 0, 0])
        bar_body_child_nodes = new_make_block_nodes(scene, bar_offset, self.size, make_parent=False, make_mesh=False)
        bar_body_mesh_node = get_scaled_mesh_node(
            scene, "res://entities/doors/rotating_bar.tscn", self.size, bar_offset
        )
        bar_body_child_nodes.append(bar_body_mesh_node)

        body_script = scene.add_ext_resource("res://entities/doors/bar_body.gd", "Script")
        bar_body_rotation_degrees = -180 if self.anchor_side == HingeSide.RIGHT else 0
        bar_body_rotation = (
            Rotation.from_euler(Axis.Z.value, bar_body_rotation_degrees, degrees=True).as_matrix().flatten()
        )
        body_node = GDNode(
            ROTATING_BAR_BODY,
            type="RigidBody",
            properties={
                "transform": make_transform(rotation=bar_body_rotation),
                "mass": 10.0,
                "script": body_script.reference,
                "entity_id": 0,
            },
        )
        for child_node in bar_body_child_nodes:
            body_node.add_child(child_node)

        support_width, support_height, support_thickness = 0.025, 0.1, 0.29
        support_size = np.array([support_width, support_height, support_thickness])
        bar_support_x = -self.size[0] * 0.8
        if self.anchor_side == HingeSide.RIGHT:
            bar_support_x = -bar_support_x
        bar_support_position = np.array([bar_support_x, 0, -support_thickness / 2.5])
        bar_support_node = only(
            new_make_block_nodes(
                scene,
                bar_support_position,
                support_size,
                parent_name="bar_support",
                make_mesh=False,
                collision_shape_disabled=True,
            )
        )
        bar_support_mesh_node = get_scaled_mesh_node(scene, "res://entities/doors/bar_support.tscn", support_size)
        bar_support_node.add_child(bar_support_mesh_node)

        bar_script = scene.add_ext_resource("res://entities/doors/rotating_bar.gd", "Script")
        group_spatial_node = GDNode(
            f"{ROTATING_BAR}_{self.entity_id}",
            "Spatial",
            properties={
                "transform": make_transform(position=self.position, rotation=self.rotation),
                "script": bar_script.reference,
                "entity_id": self.entity_id,
                "rotation_axis": self.rotation_axis.value,
                "anchor_side": self.anchor_side.value.lower(),
                "unlatch_angle": self.unlatch_angle,
            },
        )
        group_spatial_node.add_child(body_node)
        group_spatial_node.add_child(anchor_node)
        group_spatial_node.add_child(hinge_joint_node)
        group_spatial_node.add_child(bar_support_node)
        return group_spatial_node

    def get_additional_door_body_nodes(
        self, scene: GodotScene, body_size: np.ndarray, body_centroid: Point3DNP = np.array([0, 0, 0])
    ) -> List[GDNode]:
        hook_width, hook_height, hook_thickness = 0.025, 0.1, 0.2
        hook_size = np.array([hook_width, hook_height, hook_thickness])
        bar_x, bar_y, bar_z = self.position
        body_width, body_height, body_thickness = body_size
        hook_span = body_width * 0.4
        hook_distance = hook_span / 5
        hook_x_start = 0 if self.anchor_side == HingeSide.RIGHT else -hook_span
        hook_x_positions = np.arange(hook_x_start, hook_x_start + hook_span + hook_distance, hook_distance)

        hook_nodes = []
        for i, hook_x in enumerate(hook_x_positions):
            bar_hook_offset = np.array([hook_x, 0, 0])
            bar_hook_child_nodes = new_make_block_nodes(
                scene,
                bar_hook_offset,
                hook_size,
                make_parent=False,
                make_mesh=False,
                collision_shape_name=f"collision_shape_{i}",
                collision_shape_disabled=True,
            )
            bar_hook_mesh_node = get_scaled_mesh_node(
                scene,
                "res://entities/doors/bar_support.tscn",
                hook_size,
                mesh_position=bar_hook_offset,
                mesh_name=f"mesh_{i}",
            )
            hook_nodes.extend(bar_hook_child_nodes)
            hook_nodes.append(bar_hook_mesh_node)

        bar_hook_position = body_centroid + np.array([0, bar_y * 0.98, hook_thickness * 0.6])
        hook_spatial_node = GDNode(
            f"bar_hooks",
            "StaticBody",
            properties={"transform": make_transform(position=bar_hook_position)},
        )
        for hook_node in hook_nodes:
            hook_spatial_node.add_child(hook_node)
        return [hook_spatial_node]



from __future__ import annotations

from typing import Dict
from typing import Tuple

import torch
from gym import Space
from gym import spaces
from torch import Tensor
from torch import nn
from torch.nn import functional as F

from avalon.agent.common.action_model import DictActionDist
from avalon.agent.common.action_model import DictActionHead
from avalon.agent.common.models import MLP
from avalon.agent.common.types import ObservationBatch
from avalon.agent.ppo.model import PPOModel
from avalon.agent.ppo.model import mlp_init
from avalon.agent.ppo.params import PPOParams
from avalon.common.type_utils import assert_not_none


class ObservationModel(PPOModel):
    """This model is designed to match the sb3 model closely.

    Not exactly sure how close it ended up being, but that was the intent.
    """

    def __init__(self, params: PPOParams, obs_space: Space, action_space: Space) -> None:
        assert isinstance(obs_space, spaces.Dict), "We must have a dict obs space for this ObservationModel"
        assert isinstance(action_space, spaces.Dict), "We must have a dict action space for this ObservationModel"
        self.obs_space: spaces.Dict = obs_space
        self.action_space = action_space

        super().__init__()

        self.params = params
        # logger.info(f"ObservationModel with params {self.params}")

        # Concatenate image inputs, then concatenate the rest of the inputs
        _image_inputs = ("rgbd", "rgb")
        self.image_keys = sorted([x for x in obs_space.spaces.keys() if x in _image_inputs])
        encoder_output_dim = self.params.model_params.encoder_output_dim
        self.image_encoder = None
        if len(self.image_keys) > 0:
            assert len(self.image_keys) == 1, "concatenatenation of multiple images is not yet supported"
            image_key = self.image_keys[0]
            img_size = assert_not_none(obs_space[image_key].shape)[1]
            img_channels = assert_not_none(obs_space[image_key].shape)[0]
            self.image_encoder = ImpalaConvNet(
                input_channels=img_channels,
                img_dim=img_size,
                out_dim=encoder_output_dim,
                num_base_channels=self.params.model_params.num_cnn_base_channels,
            )

        # Collect remaining vectors
        self.vector_keys = sorted([x for x in obs_space.spaces.keys() if x not in _image_inputs])
        self.vector_encoder_fc = None
        if len(self.vector_keys) > 0:
            vector_dim = 0
            for key in self.vector_keys:
                space = obs_space[key]
                assert (
                    len(assert_not_none(space.shape)) == 1
                ), f"Expected non image observations to have one dim, got shape {space.shape} for {key}"
                vector_dim += assert_not_none(space.shape)[0]

            self.vector_encoder_fc = torch.nn.Linear(vector_dim, encoder_output_dim, bias=False)

            # Init to zero, keeps consistency across runs with/without proprioception
            self.vector_encoder_fc.weight.data.fill_(0.0)

        self.action_head = DictActionHead(action_space, params)
        self.policy_net = MLP(
            input_dim=encoder_output_dim,
            output_dim=self.action_head.num_inputs,
            num_layers=self.params.model_params.num_mlp_layers,
            hidden_dim=encoder_output_dim,
            activation_function=self.params.model_params.mlp_activation_fn,
        )
        # Note: this init matches baselines:ppo2 for the case of num_mlp_layers==1, but not otherwise.
        if self.params.model_params.num_mlp_layers == 1:
            mlp_init(self.policy_net.net, gain=0.01, bias=0)
        elif self.params.model_params.num_mlp_layers > 1:
            # I believe this is in ppo2. Haven't tried it yet though, been just using the default.
            # for layer in self.policy_net.net[:-1]:
            #     mlp_init(layer, gain=math.sqrt(2))
            mlp_init(self.policy_net.net[-1], gain=0.01, bias=0)  # type: ignore

        self.value_net = MLP(
            input_dim=encoder_output_dim,
            output_dim=1,
            num_layers=self.params.model_params.num_mlp_layers,
            hidden_dim=encoder_output_dim,
            activation_function=self.params.model_params.mlp_activation_fn,
        )
        # Note: this init matches baselines:ppo2 for the case of num_mlp_layers==1, but not otherwise.
        if self.params.model_params.num_mlp_layers == 1:
            mlp_init(self.value_net.net, gain=1, bias=0)

    def forward(self, obs: ObservationBatch) -> Tuple[Tensor, DictActionDist]:
        encodings = []
        if self.image_encoder:
            image_obs = self._get_image_observation(obs)
            img_encoding = self.image_encoder(image_obs)
            encodings.append(img_encoding)

        if self.vector_encoder_fc:
            vector_obs = self._get_vector_observation(obs)
            encodings.append(self.vector_encoder_fc(vector_obs))

        encoding = torch.stack(encodings, dim=0).sum(dim=0)
        batch_size = encoding.shape[0]
        value = self.value_net(encoding).reshape((batch_size,))
        action_logits = self.policy_net(encoding)
        action_dist = self.action_head(action_logits)
        return value, action_dist

    def _get_image_observation(self, obs: Dict[str, Tensor]) -> Tensor:
        return obs[self.image_keys[0]]

    def _get_vector_observation(self, obs: Dict[str, Tensor]) -> Tensor:
        """Convert the dict obs to a single vector input."""
        vector_parts = [obs[key] for key in self.vector_keys]
        vector_obs = torch.cat(vector_parts, dim=-1)
        return vector_obs


IMPALA_RES_OUT_LOOKUP = {64: 8, 84: 11, 96: 12}


class ResnetBlock(nn.Module):
    def __init__(self, n_channels: int) -> None:
        super().__init__()
        self.conv = nn.Sequential(
            nn.ReLU(),
            nn.Conv2d(in_channels=n_channels, out_channels=n_channels, kernel_size=3, stride=1, padding=1),
            nn.ReLU(),
            nn.Conv2d(in_channels=n_channels, out_channels=n_channels, kernel_size=3, stride=1, padding=1),
        )

    def forward(self, x):
        x = x + self.conv(x)
        return x


def impala_block(in_channels, out_channels):
    return nn.Sequential(
        nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=3, stride=1, padding=1),
        nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
        ResnetBlock(out_channels),
        ResnetBlock(out_channels),
    )


class ImpalaConvNet(nn.Module):
    """This has been matched carefully to the one in dreamerv2, and also openai/baselines."""

    def __init__(self, out_dim, img_dim, input_channels, num_base_channels: int = 16) -> None:
        super().__init__()
        blocks = []
        in_channels = input_channels
        for out_channels in [num_base_channels, num_base_channels * 2, num_base_channels * 2]:
            blocks.append(impala_block(in_channels, out_channels))
            in_channels = out_channels

        self.conv = nn.Sequential(*blocks)

        if img_dim not in IMPALA_RES_OUT_LOOKUP:
            raise NotImplementedError()
        res_out_dim = IMPALA_RES_OUT_LOOKUP[img_dim]
        # Note: (intentionally) leaving this with default init
        self.fc = nn.Linear(res_out_dim * res_out_dim * num_base_channels * 2, out_dim)

    def forward(self, x: torch.Tensor):
        x = self.conv(x)
        x = F.relu(x)
        x = torch.flatten(x, start_dim=1)
        x = F.relu(self.fc(x))
        return x



import json
import os
import uuid
from pathlib import Path

import boto3

from avalon.common.imports import tqdm
from avalon.common.log_utils import logger
from avalon.contrib.s3_utils import SimpleS3Client
from avalon.contrib.s3_utils import download_tar_from_s3_and_unpack
from avalon.contrib.utils import FILESYSTEM_ROOT
from avalon.contrib.utils import run_local_command
from avalon.datagen.godot_env.replay import GodotEnvReplay

GODOT_TEXTURES_PATH = Path(".") / "datagen" / "godot" / "Textures"

S3_AVALON_ERROR_BUCKET = "avalon-error-logs"


def download_godot_textures(download_path: Path = GODOT_TEXTURES_PATH) -> None:
    logger.info(f"Copying Godot textures from S3 to {download_path}")
    if not os.path.exists(str(download_path)):
        os.mkdir(download_path)

    # Create a new session to make client creation thread-safe.
    s3 = boto3.session.Session().client("s3")
    BUCKET_NAME = "godot-textures"
    bucket = boto3.resource("s3").Bucket(BUCKET_NAME)
    for s3_object in tqdm(list(bucket.objects.all())):
        path, filename = os.path.split(s3_object.key)
        filepath = download_path / filename
        if not filepath.exists():
            with open(filepath, "wb") as f:
                s3.download_fileobj(BUCKET_NAME, s3_object.key, f)


def fix_resource_paths_for_world(worlds_path: Path, target_path: Path, path_to_replace: str):
    EXCEPTED_ERRORS = {b"sed: no input files\n"}
    command = f"cd {worlds_path} && grep -rl {path_to_replace} | xargs sed -i 's+{path_to_replace}+{target_path}+g'"
    output = run_local_command(command, is_checked=False, trace_output=False)
    if output.returncode != 0 and output.stdout not in EXCEPTED_ERRORS:
        raise Exception(f"{output.stdout.decode('UTF-8')}")


def create_env_from_s3_artifacts(
    s3_key: str,
    bucket_name: str = S3_AVALON_ERROR_BUCKET,
    world_index: int = -1,
) -> GodotEnvReplay:
    s3_client = SimpleS3Client(bucket_name=bucket_name)
    output_path = Path(f"{FILESYSTEM_ROOT}/tmp/env_replay_testing/{uuid.uuid4()}/{s3_key}")
    output_path.parent.mkdir(parents=True)
    unpacked_artifacts_path = download_tar_from_s3_and_unpack(output_path, s3_client)
    return create_env_from_artifacts(unpacked_artifacts_path, world_index)


def create_env_from_artifacts(unpacked_artifacts_path: Path, world_index: int = -1) -> GodotEnvReplay:
    metadata_path = unpacked_artifacts_path / "meta.json"
    metadata = json.load(open(metadata_path))

    world_paths = unpacked_artifacts_path / "worlds"

    # note: per the convention in godot env, the world at index zero in `recent_worlds` is the oldest world
    #   to get the most recent world use `recent_worlds[-1]`
    recent_worlds = []

    for generation_id, generated_root_path in metadata["generated_root_path_by_id"].items():
        world_path = world_paths / generation_id
        fix_resource_paths_for_world(world_path, world_path, generated_root_path)
        recent_worlds.append(world_path / "main.tscn")

    return GodotEnvReplay.from_local_files(
        run_uuid=metadata["run_uuid"],
        action_path=unpacked_artifacts_path / "actions.out",
        config_path=unpacked_artifacts_path / "config.json",
        world_path=recent_worlds[world_index],
    )



from typing import Any
from typing import Dict
from typing import Mapping
from typing import Tuple
from typing import cast

import attr
import torch

from avalon.agent.common.envs import build_env
from avalon.agent.common.get_algorithm_cls import get_algorithm_cls
from avalon.agent.common.params import EnvironmentParams
from avalon.agent.common.params import Params
from avalon.agent.common.test_envs import TestEnvironmentParams
from avalon.agent.common.types import Algorithm
from avalon.agent.common.types import ParamsType
from avalon.agent.dreamer.params import DreamerParams
from avalon.agent.godot.godot_gym import GodotEnvironmentParams
from avalon.agent.ppo.params import PPOParams
from avalon.agent.train_dreamer_test_envs import DreamerTestEnvParams
from avalon.agent.train_ppo_test_envs import TestPPOParams
from avalon.contrib.testing_utils import RequestFixture
from avalon.contrib.testing_utils import fixture
from avalon.contrib.utils import make_deterministic
from avalon.datagen.world_creation.tests.params import CANONICAL_SEED

_test_environment_params = {
    "test_case1": TestEnvironmentParams(task="case1", long_episode_length=1000),
    "test_case1_continuous_action": TestEnvironmentParams(task="case1_continuous_action", long_episode_length=1000),
    "test_case2": TestEnvironmentParams(task="case2", long_episode_length=1000),
    "test_case6": TestEnvironmentParams(task="case6", long_episode_length=1000),
    "test_hybrid1": TestEnvironmentParams(task="hybrid1", long_episode_length=1000),
}


def complete_agent_params(
    agent_params: ParamsType, env_params: EnvironmentParams, seed: int, **agent_kwargs: Any
) -> ParamsType:
    make_deterministic(seed)
    dummy_env = build_env(env_params)
    agent_params = attr.evolve(
        agent_params,
        env_params=env_params,
        observation_space=dummy_env.observation_space,
        action_space=dummy_env.action_space,
        **agent_kwargs,
    )
    dummy_env.close()
    del dummy_env
    return agent_params


def _make_agent(finalized_agent_params: Params) -> Algorithm:
    assert finalized_agent_params.observation_space is not None
    assert finalized_agent_params.action_space is not None
    algorithm_cls = get_algorithm_cls(finalized_agent_params)
    return algorithm_cls(
        finalized_agent_params, finalized_agent_params.observation_space, finalized_agent_params.action_space
    )


def get_valid_agent_env_combos(seed: int) -> Mapping[Tuple[str, str], Tuple[Params, EnvironmentParams]]:
    combos: Dict[Tuple[str, str], Tuple[Params, EnvironmentParams]] = {}
    # Test envs
    for env_name, env_params in list(_test_environment_params.items()):
        combos[("ppo", env_name)] = (TestPPOParams(num_workers=1, multiprocessing=False, batch_size=50), env_params)
        combos[("dreamer", env_name)] = (
            DreamerTestEnvParams(
                num_workers=1,
                multiprocessing=False,
                prefill_eps_per_dataloader=1,
            ),
            attr.evolve(env_params),
        )

    # Common benchmark envs
    # These theoretically work for most tests, but did exhibit some non-determinism for the full training step. We
    # don't care for them that much at the moment, but may be useful in the future
    # combos[("dreamer", "atari")] = (
    #     DreamerAtariParams(num_workers=1, multiprocessing=False),
    #     DmcEnvironmentParams(suite="atari", task="atlantis", action_repeat=4, time_limit=27000),
    # )
    # combos[("dreamer", "dmc")] = (
    #     DreamerDMCParams(num_workers=1, multiprocessing=False, prefill_eps_per_dataloader=1),
    #     DmcEnvironmentParams(
    #         suite="dmc", task="walker_walk", action_repeat=2, time_limit=None, include_proprio=False, include_rgb=True
    #     ),
    # )

    # Avalon
    combos[("ppo", "avalon")] = (
        PPOParams(num_workers=1, multiprocessing=False),
        GodotEnvironmentParams(env_index=seed),
    )
    combos[("dreamer", "avalon")] = (
        DreamerParams(num_workers=1, multiprocessing=False, prefill_eps_per_dataloader=1),
        GodotEnvironmentParams(env_index=seed),
    )

    return combos


@fixture
def device_() -> torch.device:
    return torch.device("cuda")


@fixture(params=get_valid_agent_env_combos(CANONICAL_SEED).items())
def agent_and_env_(request: RequestFixture) -> Tuple[Tuple[str, str], Tuple[Params, EnvironmentParams]]:
    return cast(Tuple[Tuple[str, str], Tuple[Params, EnvironmentParams]], request.param)



import math
from typing import List
from typing import Optional
from typing import Tuple
from typing import cast

import attr
import numpy as np
from loguru import logger
from scipy import ndimage
from skimage import morphology

from avalon.common.errors import SwitchError
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.constants import IS_DEBUGGING_IMPOSSIBLE_WORLDS
from avalon.datagen.world_creation.constants import WATER_LINE
from avalon.datagen.world_creation.debug_plots import plot_terrain
from avalon.datagen.world_creation.debug_plots import plot_value_grid
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import MapFloatNP
from avalon.datagen.world_creation.types import Point2DListNP
from avalon.datagen.world_creation.types import Point2DNP
from avalon.datagen.world_creation.worlds.height_map import HeightMap


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class PointPath:
    points: np.ndarray

    def create_height_mask(
        self, map: HeightMap, path_width: float, height_bucket_count: int, gap_count: int, gap_width: float
    ) -> MapFloatNP:
        is_debugging = False

        if height_bucket_count > 0 and gap_count > 0:
            assert height_bucket_count == gap_count, "Dont support unequal numbers of buckets and gaps"

        segment_lengths = self._get_segment_lengths()
        total_length = sum(segment_lengths)

        # create each segment of the path
        mask = np.zeros_like(map.Z)
        path_length_so_far = 0.0
        for i in range(0, len(self.points) - 1):
            start_point = self.points[i]
            end_point = self.points[i + 1]
            segment_length = segment_lengths[i]
            start_val = path_length_so_far / total_length
            end_val = (path_length_so_far + segment_length) / total_length
            # gradient_mask = np.ones_like(map.Z) * start_val
            # x_delta = end_point[0] - start_point[0]
            # if abs(x_delta) > 0.0001:
            #     x_slope = (end_val - start_val) / (x_delta)
            # else:
            #     x_slope = 0
            # y_delta = end_point[1] - start_point[1]
            # if abs(y_delta) > 0.0001:
            #     y_slope = (end_val - start_val) / (y_delta)
            # else:
            #     y_slope = 0
            # x_diffs = map.X - start_point[0]
            # y_diffs = map.Y - start_point[1]
            # if x_slope > 0 and y_slope > 0:
            #     x_slope, y_slope = _normalized(np.array([x_slope, y_slope]))
            # gradient_mask += x_diffs * x_slope + y_diffs * y_slope

            # this is just a stupid way to make a gradient between the start and end point...
            center = 1000 * (start_point - end_point) + end_point
            gradient_mask = np.sqrt(map.get_dist_sq_to(center))
            gradient_mask -= gradient_mask[map.point_to_index(start_point)]
            end_dist_val = gradient_mask[map.point_to_index(end_point)]
            if end_dist_val > 0.00001:
                gradient_mask /= end_dist_val

                gradient_mask = np.clip(gradient_mask, start_val, end_val)
                line_seg_distances = map.get_lineseg_distances(start_point, end_point)
                close_enough = line_seg_distances < (path_width / 2.0)
                mask[close_enough] = gradient_mask[close_enough]

                if is_debugging:
                    logger.debug("Path segment:")
                    logger.debug((start_val, end_val))
                    logger.debug((start_point, end_point))
                    plot_value_grid(gradient_mask, "Path construction gradient")
                    plot_value_grid(line_seg_distances, "Path construction line seg distances")
                    plot_value_grid(close_enough, "Path construction close enough")
                    plot_value_grid(
                        mask,
                        "Path construction of height mask",
                        markers=[map.point_to_index(x) for x in [start_point, end_point]],
                    )

            # if gap_count > 0:
            #     gaps = gap_count * min(1, round(segment_length / total_length))
            # else:
            #     gaps = 0
            # gap_locations = asdfsda
            # t = asdfdsaf
            # gap_mask = asdfdsaf
            # for gap_location in gap_locations:
            #     gap_start = gap_location - gap_width / 2.0
            #     gap_end = gap_location + gap_width / 2.0
            #     gap_mask[np.logical_and(t > gap_start, t < gap_end)] = True
            # solid_mask = np.logical_not(gap_mask)

            path_length_so_far += segment_length

        # with the requested number of gaps for that segment
        # create the composite path mask
        # bucket the values
        # if height_bucket_count > 0:
        #     mask = asdfdsa

        return mask

    def get_length(self) -> float:
        return sum(self._get_segment_lengths())

    def _get_segment_lengths(self) -> Tuple[float, ...]:
        segments = []
        for i in range(1, len(self.points)):
            dist = float(np.linalg.norm(self.points[i] - self.points[i - 1]))
            segments.append(dist)
        return tuple(segments)


class PathFailure(Exception):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HeightPath:
    # this is the width of the full path, ie, half of this width from the middle to each edge
    width: float = 1.0
    extra_point_count: int = 0
    is_path_climbable: bool = False
    is_height_affected: bool = True
    is_solution_flattened: bool = False
    is_path_restricted_to_land: bool = False
    flattening_mode: str = "midpoint"
    is_chasm_bottom_flattened: bool = False
    is_detailed: bool = True
    is_outside_edge_unclimbable: bool = False
    min_desired_2d_length: Optional[float] = None
    max_desired_2d_length: Optional[float] = None
    height_bucket_count: int = 0
    gap_count: int = 0
    gap_width: float = 0.0
    max_retries: int = 10
    ocean_radius: float = 20.0
    is_path_failure_allowed: bool = False
    is_debug_graph_printing_enabled: bool = False

    def apply(
        self,
        rand: np.random.Generator,
        map: HeightMap,
        is_climbable: MapBoolNP,
        is_detail_important: MapBoolNP,
        start_point: Point2DNP,
        end_point: Point2DNP,
        obstacle_weight_map: MapFloatNP,
        inner_solution_mask: MapBoolNP,
        outer_solution_mask: MapBoolNP,
    ):

        full_mask = np.logical_or(np.logical_or(inner_solution_mask, outer_solution_mask), obstacle_weight_map > 0.0)
        if self.is_debug_graph_printing_enabled:
            plot_value_grid(full_mask, "Full mask", markers=[map.point_to_index(x) for x in [start_point, end_point]])
            # plot_value_grid(inner_solution_mask)
            # plot_value_grid(outer_solution_mask)

        start_to_end_2d_dist = float(np.linalg.norm(start_point - end_point))
        if self.min_desired_2d_length is None:
            min_len = start_to_end_2d_dist
        else:
            min_len = self.min_desired_2d_length
        if self.max_desired_2d_length is None:
            max_len = 3 * min_len
        else:
            max_len = self.max_desired_2d_length

        # the basic path. Used if cannot make path
        path = PointPath(np.array([start_point, end_point]))
        # different cases for each:
        if self.extra_point_count == 0:
            # if there are 0 extra points, just a straight shot
            pass
        elif self.extra_point_count == 1:
            one_point_path = self._create_one_point_path(
                rand, map, full_mask, start_point, end_point, min_len, max_len
            )
            if one_point_path is not None:
                path = one_point_path
            elif not self.is_path_failure_allowed:
                raise PathFailure()
            else:
                map.log_simplicity_warning("Reduce from 1 extra point path to direct path")
        elif self.extra_point_count == 2:
            two_point_path = self._create_two_point_path(
                rand, map, full_mask, start_point, end_point, min_len, max_len
            )
            if two_point_path is not None:
                path = two_point_path
            elif not self.is_path_failure_allowed:
                raise PathFailure()
            else:
                map.log_simplicity_warning("Reduce from 1 extra point path to direct path")
        else:
            raise SwitchError(f"Only support paths with <= 2 points, sorry")

        # figure out the height differences
        start_height = map.get_rough_height_at_point(start_point)
        end_height = map.get_rough_height_at_point(end_point)
        height_delta = end_height - start_height
        # logger.debug(f"{height_delta=}")

        if self.is_solution_flattened or self.is_chasm_bottom_flattened:
            assert not self.is_height_affected, "Flattening and affecting height are mutually exclusive"
            assert self.extra_point_count == 0, "Only support flattening single segment paths"
            assert not self.is_outside_edge_unclimbable, "Doesnt make sense when flattening"
            assert not self.is_path_climbable, "Doesnt make sense when flattening"
            path_distances = map.get_lineseg_distances(start_point, end_point)
            radius = self.width / 2.0
            path_mask = path_distances < radius

            solution_only_mask = np.logical_or(inner_solution_mask, outer_solution_mask)

            if self.is_debug_graph_printing_enabled:
                plot_value_grid(path_mask, "Path Mask")

            if self.is_path_restricted_to_land:
                if np.any(np.logical_and(np.logical_not(map.get_land_mask()), path_mask)):
                    if IS_DEBUGGING_IMPOSSIBLE_WORLDS:
                        plot_terrain(
                            map.Z, "Path endpoints", markers=[map.point_to_index(x) for x in [start_point, end_point]]
                        )
                        plot_value_grid(path_mask, "Path")
                    raise ImpossibleWorldError("Path ended up in the ocean, not good. Try again plz.")

            # even if we're not restricted to land, we cannot affect the water
            path_mask = np.logical_and(path_mask, full_mask)

            # actually apply the flattening to the solution parts
            if self.is_solution_flattened:
                if self.flattening_mode == "midpoint":
                    height = (start_height + end_height) / 2.0
                elif self.flattening_mode == "min":
                    height = min([start_height, end_height])
                elif self.flattening_mode == "max":
                    height = max([start_height, end_height])
                else:
                    raise SwitchError(f"Unsupported {self.flattening_mode}")
                falloff_mask = np.logical_and(path_distances < self.width, solution_only_mask)
                mixing = np.clip(((self.width - path_distances[falloff_mask]) / self.width) * 2, 0, 1)
                map.Z[falloff_mask] = map.Z[falloff_mask] * (1.0 - mixing) + height * mixing

            # and apply the flattening to the obstacle parts
            if self.is_chasm_bottom_flattened:
                mid_point = (start_point + end_point) / 2.0
                height = map.get_rough_height_at_point(mid_point)
                falloff_mask = np.logical_and(path_distances < self.width, obstacle_weight_map >= 1.0)
                mixing = np.clip(((self.width - path_distances[falloff_mask]) / self.width) * 2, 0, 1)
                map.Z[falloff_mask] = map.Z[falloff_mask] * (1.0 - mixing) + height * mixing
        else:
            height_mask = path.create_height_mask(
                map, self.width, self.height_bucket_count, self.gap_count, self.gap_width
            )
            path_mask = np.logical_and(height_mask > 0, height_mask < 1)

            if self.is_path_restricted_to_land:
                if np.any(np.logical_and(np.logical_not(map.get_land_mask()), path_mask)):
                    if IS_DEBUGGING_IMPOSSIBLE_WORLDS:
                        plot_terrain(
                            map.Z, "Path endpoints", markers=[map.point_to_index(x) for x in [start_point, end_point]]
                        )
                        plot_value_grid(path_mask, "Path")
                    raise ImpossibleWorldError("Path ended up in the ocean, not good. Try again plz.")

            # only apply the part that overlaps with our solution zones
            path_mask = np.logical_and(path_mask, full_mask)

            if self.is_debug_graph_printing_enabled:
                plot_value_grid(path_mask, "Path Mask")

            # actually apply the path
            if self.is_height_affected:
                map.Z[path_mask] = height_mask[path_mask] * height_delta + start_height
            if self.is_path_climbable:
                is_climbable[path_mask] = True
            if self.is_outside_edge_unclimbable:
                expanded_mask = ndimage.binary_dilation(path_mask, structure=morphology.disk(5))
                edge_mask = np.logical_and(expanded_mask, np.logical_not(path_mask))
                edge_mask = ndimage.binary_dilation(edge_mask, structure=morphology.disk(1))
                edge_mask = np.logical_and(edge_mask, obstacle_weight_map > 0.0)
                is_climbable[edge_mask] = False
                is_detail_important[edge_mask] = True

        if self.is_detailed:
            is_detail_important[path_mask] = True

        if self.is_debug_graph_printing_enabled:
            plot_value_grid(is_climbable, "Is climbable after path")
            plot_value_grid(is_detail_important, "Is detail after path")

    def _create_one_point_path(
        self,
        rand: np.random.Generator,
        map: HeightMap,
        full_mask: MapBoolNP,
        start_point: Point2DNP,
        end_point: Point2DNP,
        min_len: float,
        max_len: float,
    ) -> Optional[PointPath]:
        # try a bunch of splits for the desired distance
        start_dist_sq = map.get_dist_sq_to(start_point)
        end_dist_sq = map.get_dist_sq_to(end_point)
        for mask in self._masks_to_try(map, full_mask, start_dist_sq, end_dist_sq):
            for i in range(self.max_retries):
                overall_len = rand.uniform(min_len, max_len)
                path = _make_one_point_path(rand, map, mask, overall_len, start_point, end_point, self.width)
                if path is not None:
                    return path
        return None

    def _create_two_point_path(
        self,
        rand: np.random.Generator,
        map: HeightMap,
        full_mask: MapBoolNP,
        start_point: Point2DNP,
        end_point: Point2DNP,
        min_len: float,
        max_len: float,
    ) -> Optional[PointPath]:
        start_dist_sq = map.get_dist_sq_to(start_point)
        end_dist_sq = map.get_dist_sq_to(end_point)
        max_len_in_one_point = (max_len - min_len) * 0.6
        for mask in self._masks_to_try(map, full_mask, start_dist_sq, end_dist_sq):
            for i in range(self.max_retries):
                overall_len = rand.uniform(min_len, min_len + max_len_in_one_point * rand.uniform())
                start_len = overall_len * rand.uniform(0.1, 0.45)
                end_len = start_len
                mid_points = _get_circle_intersections(start_point, start_len, end_point, end_len)
                if len(mid_points) == 2:
                    rand.shuffle(mid_points)
                    path = PointPath(np.array([start_point, *mid_points, end_point]))
                    if _check_full_path(path, map, mask, self.width):
                        if min_len < path.get_length() < max_len:
                            return path
                        else:
                            logger.debug("Rejected because path length")

                    # fine, whatever, let's just try again
        # if we've tried a bunch and still failed, bail
        return None

    # allows us to try again cutting a little bit into the ocean if strictly necessary to find a working path
    def _masks_to_try(
        self, map: HeightMap, full_mask: MapBoolNP, start: MapFloatNP, end: MapFloatNP
    ) -> List[MapBoolNP]:
        if self.ocean_radius <= 0.0:
            return [full_mask]
        alt_mask = full_mask.copy()
        ocean_radius_sq = self.ocean_radius * self.ocean_radius
        within_radius = np.logical_or(start < ocean_radius_sq, end < ocean_radius_sq)
        alt_mask[np.logical_and(map.Z < WATER_LINE, within_radius)] = True
        return [full_mask, alt_mask]


def _make_one_point_path(
    rand: np.random.Generator,
    map: HeightMap,
    mask: np.ndarray,
    overall_len: float,
    start_point: Point2DNP,
    end_point: Point2DNP,
    width: float,
) -> Optional[PointPath]:
    start_len = overall_len * 0.5
    end_len = start_len
    for mid_point in _get_circle_intersections(start_point, start_len, end_point, end_len):
        is_start_path_clear = _check_path(map, mask, start_point, mid_point, width, shrink_start=True)
        is_end_path_clear = _check_path(map, mask, start_point, mid_point, width, shrink_end=True)
        if is_start_path_clear and is_end_path_clear:
            return PointPath(np.array([start_point, mid_point, end_point]))
    return None


def _check_full_path(path: PointPath, map: HeightMap, mask: MapBoolNP, width: float) -> bool:
    # plot_value_grid(mask, markers=[map.point_to_index(x) for x in path.points])

    for i in range(1, len(path.points)):
        start = path.points[i - 1]
        end = path.points[i]
        if not _check_path(map, mask, start, end, width, shrink_start=i == 1, shrink_end=i == (len(path.points) - 1)):
            return False
        # # ensure that non of the other points are co-linear
        # other_points = list(path.points[: i - 1]) + list(path.points[i + 1 :])
        # for point in other_points:
        #     if abs(convenient_signed_line_distance(point, start, end)) < 0.5 * width:
        #         logger.debug("Rejected because colinear")
        #         return False
    return True


def _check_path(
    map: HeightMap,
    mask: MapBoolNP,
    start_point: Point2DNP,
    end_point: Point2DNP,
    width: float,
    shrink_start: bool = False,
    shrink_end: bool = True,
) -> bool:
    if shrink_start:
        points_2d = np.stack([map.X, map.Y], axis=2)
        path_mask = _make_triangular_mask(points_2d, start_point, end_point, width)
    elif shrink_end:
        points_2d = np.stack([map.X, map.Y], axis=2)
        path_mask = _make_triangular_mask(points_2d, end_point, start_point, width)
    else:
        # shrink the points in so they dont intersect at the ends
        start_to_end_dist = np.linalg.norm(start_point - end_point)
        if start_to_end_dist < width:
            midpoint = (start_point + end_point) / 2.0
            path_mask = map.get_dist_sq_to(midpoint) < width**2
        else:
            start_to_end_vec = (end_point - start_point) / start_to_end_dist
            start_point = start_point + start_to_end_vec * (width / 2.0)
            end_point = end_point + -1.0 * start_to_end_vec * (width / 2.0)
            line_seg_distances = map.get_lineseg_distances(start_point, end_point)
            path_mask = line_seg_distances < width / 2.0

    # plot_value_grid(path_mask, "mask", markers=[map.point_to_index(x) for x in [start_point, end_point]])

    if np.any(np.logical_and(path_mask, np.logical_not(mask))):
        # plot_value_grid(mask, "mask", markers=[map.point_to_index(x) for x in [start_point, end_point]])
        # plot_value_grid(line_seg_distances < width, "line_seg_distances")
        # plot_value_grid(np.logical_and(line_seg_distances < width, np.logical_not(mask)), "failure points")
        # logger.debug("Rejected because not in mask")
        return False
    return True


def _make_triangular_mask(points_2d: Point2DListNP, start: Point2DNP, end: Point2DNP, width: float) -> MapBoolNP:
    start_to_end_dist = np.linalg.norm(start - end)
    start_to_end_vec = (end - start) / start_to_end_dist
    rot_vec = np.array([-start_to_end_vec[1], start_to_end_vec[0]])
    e1 = end + rot_vec * width / 2.0
    e2 = end + -1.0 * rot_vec * width / 2.0

    triangle_mask = np.logical_and(
        np.logical_and(
            _bulk_signed_line_distance(points_2d, start, e1) > 0, _bulk_signed_line_distance(points_2d, e1, e2) > 0
        ),
        _bulk_signed_line_distance(points_2d, e2, start) > 0,
    )

    # plot_value_grid(triangle_mask)

    return cast(MapBoolNP, triangle_mask)


def _bulk_signed_line_distance(points: np.ndarray, a: Point2DNP, b: Point2DNP) -> np.ndarray:
    ab_dist = math.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)
    return cast(
        np.ndarray, (((b[0] - a[0]) * (a[1] - points[:, :, 1])) - ((a[0] - points[:, :, 0]) * (b[1] - a[1]))) / ab_dist
    )


def _get_circle_intersections(
    center_a: Point2DNP, radius_a: float, center_b: Point2DNP, radius_b: float
) -> List[Point2DNP]:
    result = _get_intersections(center_a[0], center_a[1], radius_a, center_b[0], center_b[1], radius_b)
    if result is None:
        return []
    x1, y1, x2, y2 = result
    if abs(x1 - x2) < 0.001 and abs(y1 - y2) < 0.001:
        return [np.array([x1, y1])]
    return [np.array([x1, y1]), np.array([x2, y2])]


def _get_intersections(
    x0: float, y0: float, r0: float, x1: float, y1: float, r1: float
) -> Optional[Tuple[float, float, float, float]]:
    # circle 1: (x0, y0), radius r0
    # circle 2: (x1, y1), radius r1

    d = math.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)

    # non intersecting
    if d > r0 + r1:
        return None
    # One circle within other
    if d < abs(r0 - r1):
        return None
    # coincident circles
    if d == 0 and r0 == r1:
        return None
    else:
        a = (r0**2 - r1**2 + d**2) / (2 * d)
        h = math.sqrt(r0**2 - a**2)
        x2 = x0 + a * (x1 - x0) / d
        y2 = y0 + a * (y1 - y0) / d
        x3 = x2 + h * (y1 - y0) / d
        y3 = y2 - h * (x1 - x0) / d

        x4 = x2 - h * (y1 - y0) / d
        y4 = y2 + h * (x1 - x0) / d

        return (x3, y3, x4, y4)



from pathlib import Path
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Type

import attr
import numpy as np
from scipy import stats

from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import DEFAULT_SAFETY_RADIUS
from avalon.datagen.world_creation.constants import ITEM_FLATTEN_RADIUS
from avalon.datagen.world_creation.constants import WEAPON_HEIGHT_OFFSET
from avalon.datagen.world_creation.entities.animals import ALL_PREY_CLASSES
from avalon.datagen.world_creation.entities.animals import Animal
from avalon.datagen.world_creation.entities.animals import Prey
from avalon.datagen.world_creation.entities.entity import Entity
from avalon.datagen.world_creation.entities.tools.placeholder import Placeholder
from avalon.datagen.world_creation.entities.tools.weapons import LargeRock
from avalon.datagen.world_creation.entities.tools.weapons import LargeStick
from avalon.datagen.world_creation.entities.tools.weapons import Rock
from avalon.datagen.world_creation.entities.tools.weapons import Stick
from avalon.datagen.world_creation.entities.tools.weapons import Weapon
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_categorical_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import create_inner_placeholder_solution
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations

TotalWeaponValue = Literal[4, 3, 2, 1]


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class HuntTaskConfig(TaskConfig):
    # how large of a world to create. Different bounds depending on if inside a pit or outside of the pit.
    # Note that it is much easier when the prey is in a pit because they cannot escape and ricochets can hit the prey
    inside_pit_dist_easy: float = 5.0
    inside_pit_dist_hard: float = 20.0
    outside_pit_dist_easy: float = 10.0
    outside_pit_dist_hard: float = 40.0
    # how big the pit should be. It will grow on harder difficulties up to the max possible in the world given the
    # pit distance defined above
    pit_size_min: float = 2.0
    # how steep to make the pit. The length is uniform random between these two values.
    pit_incline_length_min: float = 0.5
    pit_incline_length_max: float = 1.75
    # how high the pit should be. Is a bit weirdly parameterized because really the easiest thing is if the prey is just
    # baarely unable to escape, but not in such a deep pit that it makes it hard to throw stuff at it
    shallow_pit_depth_easy: float = 1.0
    shallow_pit_depth_hard: float = 2.0
    deep_pit_depth_easy: float = 3.0
    deep_pit_depth_hard: float = 2.0
    pit_depth_std_dev: float = 0.25
    # these parameters control which weapons will be spawned
    # the probability that any given weapons will be a rock (vs a stick)
    rock_probability: float = 0.5
    # some weapons are more useful than others. They are spawned until enough "value" has been spawned
    weapon_value_easy: float = 5.0
    weapon_value_hard: float = 2.0
    weapon_value_std_dev: float = 0.5
    # there are large variants of weapons, which do more damage. They become less likely to spawn at higher difficulties
    large_weapon_probability_easy: float = 0.8
    large_weapon_probability_hard: float = 0.2


@attr.define
class ForceWeapons:
    # increments of 0.5
    weapon_value: Optional[TotalWeaponValue] = None
    weapon_radius: Optional[float] = None
    rock_probability: Optional[float] = None
    large_weapon_probability: Optional[float] = None
    desired_goal_distance: Optional[float] = None


@attr.define
class ForceHunt(ForceWeapons):
    prey: Optional[Type[Prey]] = None
    is_prey_in_pit: Optional[bool] = None
    is_pit_shallow: Optional[bool] = None
    pit_depth: Optional[float] = None


def generate_hunt_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: HuntTaskConfig = HuntTaskConfig(),
    _FORCED: Optional[ForceHunt] = None,
) -> None:
    world, locations, difficulty = create_hunt_obstacle(
        rand, difficulty, export_config, _forced_hunt=_FORCED, task_config=task_config
    )
    world, locations = world.end_height_obstacles(locations, is_accessible_from_water=True)
    world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)
    export_world(output_path, rand, world)


def create_hunt_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    _forced_hunt: Optional[ForceHunt] = None,
    task_config: HuntTaskConfig = HuntTaskConfig(),
) -> Tuple[World, WorldLocations, float]:
    if _forced_hunt is None:
        _forced_hunt = ForceHunt()

    is_prey_in_pit, difficulty = select_boolean_difficulty(difficulty, rand, _FORCED=_forced_hunt.is_prey_in_pit)
    prey_type, output_difficulty = select_categorical_difficulty(
        ALL_PREY_CLASSES, difficulty, rand, _FORCED=_forced_hunt.prey
    )

    if is_prey_in_pit:
        # this isn't exactly distance... because we reset the spawn position to be at the ledge below
        #  so we will actually end up closer
        desired_goal_distance = scale_with_difficulty(
            difficulty,
            task_config.inside_pit_dist_easy,
            task_config.inside_pit_dist_hard,
            _FORCED=_forced_hunt.desired_goal_distance,
        )
    else:
        # this WILL be how far away the prey is
        desired_goal_distance = scale_with_difficulty(
            difficulty,
            task_config.outside_pit_dist_easy,
            task_config.outside_pit_dist_hard,
            _FORCED=_forced_hunt.desired_goal_distance,
        )

    prey_offset = prey_type(position=np.array([0.0, 0.0, 0.0])).get_offset()
    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_distance, 0.5), rand, difficulty, export_config, constraint, prey_offset
    )

    pit_incline_delta = task_config.pit_incline_length_max - task_config.pit_incline_length_min
    pit_incline_distance = task_config.pit_incline_length_min + pit_incline_delta * rand.uniform()
    max_possible_outer_radius, _warnings = world.get_critical_distance(
        locations, pit_incline_distance + task_config.pit_size_min
    )

    prey = prey_type(position=locations.goal)

    weapons, output_difficulty = select_weapons(
        rand,
        difficulty,
        is_rock_necessary=is_flying_prey_present([prey]),
        rock_probability=_forced_hunt.rock_probability or task_config.rock_probability,
        weapon_value_easy=task_config.weapon_value_easy,
        weapon_value_hard=task_config.weapon_value_hard,
        weapon_value_std_dev=task_config.weapon_value_std_dev,
        large_weapon_probability_easy=task_config.large_weapon_probability_easy,
        large_weapon_probability_hard=task_config.large_weapon_probability_hard,
        _FORCED_WEAPON_VALUE=_forced_hunt.weapon_value,
        _FORCED_LARGE_WEAPON_RATE=_forced_hunt.large_weapon_probability,
    )

    weapon_radius = difficulty_variation(3.0, 6.0, rand, difficulty, _FORCED=_forced_hunt.weapon_radius)

    # if there is no pit, or the world is too small for a pit, fine, just stick some weapons and predator there
    if max_possible_outer_radius is None or not is_prey_in_pit:
        world = world.add_item(prey, reset_height_offset=prey.get_offset())
        midway = locations.spawn + (locations.goal - locations.spawn) / rand.uniform(1.5, 2.5)
        mid_point = to_2d_point(midway)
        for _ in range(len(weapons)):
            point = world.get_random_point_for_weapon_or_predator(rand, mid_point, weapon_radius, locations.island)
            if point is None:
                raise ImpossibleWorldError("No place for any weapons")
            position = np.array([point[0], 0.0, point[1]])
            world = world.add_item(Placeholder(position=position), reset_height_offset=WEAPON_HEIGHT_OFFSET)
        world = world.replace_weapon_placeholders(weapons, locations.island, ITEM_FLATTEN_RADIUS)

        new_locations = locations
    else:
        # can be increased up to max_possible_outer_radius
        # makes the task harder because then the pit is bigger
        extra_pit_size = scale_with_difficulty(difficulty, task_config.pit_size_min, max_possible_outer_radius)
        outer_radius = max_possible_outer_radius - extra_pit_size

        # Smaller walls are potentially harder because the prey can escape, but whatever
        is_pit_shallow, difficulty = select_boolean_difficulty(difficulty, rand, _FORCED=_forced_hunt.is_pit_shallow)
        if is_pit_shallow:
            pit_range = task_config.shallow_pit_depth_easy, task_config.shallow_pit_depth_hard
        else:
            pit_range = task_config.deep_pit_depth_easy, task_config.deep_pit_depth_hard
        pit_depth = -normal_distrib_range(
            pit_range[0], pit_range[1], task_config.pit_depth_std_dev, rand, difficulty, _FORCED=_forced_hunt.pit_depth
        )

        weapon_placeholders = attr.evolve(
            create_inner_placeholder_solution(
                count=len(weapons),
                offset=WEAPON_HEIGHT_OFFSET,
                randomization_dist=weapon_radius,
            ),
            inside_item_radius=ITEM_FLATTEN_RADIUS,
        )

        ring_config = make_ring(
            rand,
            difficulty,
            world,
            locations,
            constraint=constraint,
            gap_distance=0.0,
            height=pit_depth,
            traversal_width=normal_distrib_range(10.0, 1.0, 1.0, rand, difficulty),
            inner_traversal_length=pit_incline_distance,
            is_single_obstacle=True,
            inner_solution=weapon_placeholders,
            # always centered around the goal to ensure that it is a pit
            probability_of_centering_on_spawn=0.0,
            outer_traversal_length=outer_radius,
            # disables extra large circles so we can make really small pits
            max_additional_radius_multiple=1.0,
            # mostly this way so that YOU dont get stuck down there.
            # flying prey can always escape though
            is_inside_climbable=True,
        )
        world = world.add_height_obstacle(rand, ring_config, locations.island)
        world = world.replace_weapon_placeholders(weapons, locations.island, ITEM_FLATTEN_RADIUS)

        world = world.add_item(prey, reset_height_offset=prey.get_offset())

        # reset the spawn location to the bottom edge of the pit so we are right next to the prey
        weapon_positions = [x.position for x in world.items if is_weapon(x)]
        if len(weapon_positions) > 0:
            average_weapon_position = np.array(weapon_positions).mean(axis=0)
            new_locations = attr.evolve(locations, spawn=average_weapon_position)
        else:
            point = world.get_random_point_for_weapon_or_predator(
                rand,
                to_2d_point(locations.goal),
                extra_pit_size + DEFAULT_SAFETY_RADIUS + pit_incline_distance,
                locations.island,
            )
            if point is None:
                new_locations = locations
            else:
                position = np.array([point[0], 0.0, point[1]])
                new_locations = attr.evolve(locations, spawn=position)
    return world, new_locations, difficulty


def select_weapons(
    rand: np.random.Generator,
    difficulty: float,
    is_rock_necessary: bool,
    rock_probability: float,
    weapon_value_easy: float,
    weapon_value_hard: float,
    weapon_value_std_dev: float,
    large_weapon_probability_easy: float,
    large_weapon_probability_hard: float,
    _FORCED_WEAPON_VALUE: Optional[TotalWeaponValue] = None,
    _FORCED_LARGE_WEAPON_RATE: Optional[float] = None,
) -> Tuple[List[Type[Weapon]], float]:

    # Rocks/projectiles are harder to use and you lose them, so we give each weapon a "value" instead of having fixed counts
    weapon_value = int(
        normal_distrib_range(
            weapon_value_easy + 0.49, weapon_value_hard - 0.49, weapon_value_std_dev, rand, difficulty
        )
    )

    # Large weapons have the same value as their counterparts for now
    large_weapon_rate = scale_with_difficulty(
        difficulty, large_weapon_probability_easy, large_weapon_probability_hard, _FORCED=_FORCED_LARGE_WEAPON_RATE
    )

    weapons: List[Type[Weapon]] = []

    if is_rock_necessary:
        weapons.append(Rock)
        weapon_value -= Rock.WEAPON_VALUE
    weapon: Type[Weapon]
    while weapon_value > 0:
        if weapon_value < 1 or rand.uniform() < rock_probability:
            weapon = Rock
        else:
            weapon = Stick
        if rand.uniform() < large_weapon_rate:
            weapon = LargeRock if weapon is Rock else Stick
        weapon_value -= weapon.WEAPON_VALUE
        weapons.append(weapon)

    return weapons, difficulty


def is_flying_prey_present(animals: List[Animal]):
    return any(animal for animal in animals if isinstance(animal, Prey) and animal.is_flying)


def is_weapon(item: Entity):
    return isinstance(item, (Rock, Stick, Placeholder, LargeRock, LargeStick))



from typing import Tuple

import attr
import numpy as np

from avalon.datagen.world_creation.types import MapFloatNP


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HarmonicsConfig:
    harmonics: Tuple[int, ...]
    weights: Tuple[float, ...]
    # just used for debugging
    is_deterministic: bool = False


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class EdgeConfig:
    # edges are scaled based on their radius so that the size of the noise is constant
    # this factor changes that automatic scaling
    # a higher number will effectively scale up a smaller circle by that much
    scale: float = 1.0
    # noise of 0.0 means "no noise, smooth line", noise of 1.0 means "very noisy"
    # going much beyond 1.0 is probably a bad idea
    noise: float = 0.5
    # 1.0 means a perfect circle, 0.0 means weirdly blobby
    circularity: float = 0.5

    def to_harmonics(self, radius: float) -> HarmonicsConfig:
        noise = self.noise**0.5
        base_radius = 20.6
        base_harmonic_count = 20 * noise
        scale_factor = (radius / base_radius) * (1.0 / self.scale)
        noise_harmonics = list(range(1, round(base_harmonic_count * scale_factor**0.5)))
        scaled_harmonics = tuple([round(2 * scale_factor) + i for i in noise_harmonics])
        noise_weights = tuple([noise * 0.75 / x for x in scaled_harmonics])

        shape_factor = (1.0 - self.circularity) * 1.75
        shape_weights = tuple(x * shape_factor for x in [0.1, 0.3, 0.2, 0.1, 0.05])
        shape_harmonics = (1, 2, 3, 4, 5)

        return HarmonicsConfig(
            harmonics=scaled_harmonics + shape_harmonics,
            weights=noise_weights + shape_weights,
        )


def create_harmonics(
    rand: np.random.Generator, theta: MapFloatNP, config: HarmonicsConfig, is_normalized: bool
) -> MapFloatNP:
    variation = np.zeros_like(theta)
    for harmonic, weight in zip(config.harmonics, config.weights):
        # noise = rand.normal()
        noise: float = rand.uniform(-1, 1)
        if config.is_deterministic:
            noise = 1.0
        variation += np.sin(harmonic * theta) * ((noise) * weight)
        # noise = rand.normal()
        noise = rand.uniform(-1, 1)
        if config.is_deterministic:
            noise = 1.0
        variation += np.cos(harmonic * theta) * ((noise) * weight)
    if is_normalized:
        variation_min = np.min(variation)
        variation = (variation - variation_min) / (np.max(variation) - variation_min)
    return variation



import pathlib
import pickle
import random
import time
from typing import Iterator

import attr
import numpy as np
from loguru import logger
from torch.utils.data import IterableDataset
from tree import map_structure

from avalon.agent.common.types import SequenceData
from avalon.agent.dreamer.params import OffPolicyParams

# TODO: consider redoing this with memory-mapped np arrays as the storage.
# Can load slices of them from disk (automatically cached by kernel as needed) without reading the whole file
# or having to keep everything in memory.
# Would be a bit tricky to implement though since we have a bunch of arrays per episode.
# I guess not so many that we wouldn't just have a file for each, although some files would be p small.


def partition_file(filename: str, i: int, num_workers: int) -> bool:
    """Given a filename of the form blah-UUID4-blah, like 20220608T113534-bfc1430dab5245dbac6d0a34f906244f-49.npz,
    decide if it is assigned to this worker."""
    # TODO: this janky method doesn't seem to make quiiiite an even split. But so far it's good enough.
    x = ord(filename.split("-")[1][0])
    return x % num_workers == i


WORKER_ID = 0


def worker_init_fn(i: int) -> None:
    """We want an incrementing id for each dataloader; I can't find a cleaner way than this."""
    global WORKER_ID
    WORKER_ID = i


class ReplayDataset(IterableDataset):
    """A dataset that loads episodes from disk and returns random fragments from those episodes.

    Episodes are cached in memory for higher throughput.
    The full dataset is split among the `params.num_dataloader_workers` workers to save memory.
    """

    def __init__(self, params: OffPolicyParams, storage_dir: str, update_interval: int = 100) -> None:
        self.params = params
        self.cache: dict[str, SequenceData] = {}
        # For O(1) key sampling. Must manually maintain that set(self.storage.keys()) == set(self.storage_keys)
        self.cache_keys: list[str] = []
        self.storage_dir = pathlib.Path(storage_dir).expanduser()
        self.storage_dir.mkdir(parents=True, exist_ok=True)
        self.time_since_last_update = None
        self.update_interval = update_interval

        # Clear any old files in this folder, without deleting it (as other processes may already be accessing it).
        for f in self.storage_dir.iterdir():
            if f.is_file():
                f.unlink()

    def load_into_cache(self) -> None:
        """Update the cache to match the state of the disk folder."""
        global WORKER_ID
        disk_filenames = [f.name for f in sorted(pathlib.Path(self.storage_dir).iterdir()) if f.is_file()]
        # Shuffle so we're not loading a biased part of the dataset first
        random.shuffle(disk_filenames)
        loaded = 0
        for filename in disk_filenames:
            # only load our slice of the dataset
            if not partition_file(filename, WORKER_ID, self.params.num_dataloader_workers):
                continue
            if filename not in self.cache:
                try:
                    with open(self.storage_dir / filename, "rb") as f:
                        episode: SequenceData = pickle.load(f)
                except Exception as e:
                    logger.warning(f"Could not load episode: {e}")
                    continue
                self.cache[filename] = episode
                self.cache_keys.append(filename)
                loaded += 1
                # # When we open an existing dataset, let's not load it all in upfront, which would block training..
                # if loaded > 500:
                #     logger.info("loaded 500 eps, continuing")
                #     break
        disk_filenames_set = set(disk_filenames)
        for filename in self.cache_keys:
            if filename not in disk_filenames_set:
                del self.cache[filename]
                self.cache_keys.remove(filename)

    def get_trajectory(self):
        """Randomly sample a single trajectory fragment from the dataset.

        Samples from episodes randomly. May not be ideal if episodes are of widely varying length;
        may be preferrable to weight samples by length.
        """
        global WORKER_ID
        if self.time_since_last_update is None or self.time_since_last_update > self.update_interval:
            # TODO: alter this somehow so that all dataloaders aren't updating their caches at the same time.
            self.load_into_cache()
            self.time_since_last_update = 0
        else:
            self.time_since_last_update += 1

        while len(self.cache) < self.params.prefill_eps_per_dataloader:
            logger.debug("waiting for samples")
            time.sleep(5)
            self.load_into_cache()

        while True:
            # Can relax this once we add the is_first handling and handle this properly in algos.
            # This codepath is otherwise ready for variable length episodes.
            assert self.params.min_fragment_len == self.params.max_fragment_len

            episode = self.cache[random.choice(self.cache_keys)]

            total = len(episode.done)
            assert total >= self.params.min_fragment_len
            length = total
            length = min(length, self.params.max_fragment_len)
            # Randomize length to avoid all chunks ending at the same time in case the
            # episodes are all of the same length.
            # TODO: why?
            # This will only do something if we have frames to spare
            # (and tbh could be improved such that with 55 frames available and a min of 50, we apply randomness more often than 10% of the time).
            length -= np.random.randint(self.params.min_fragment_len)
            length = max(self.params.min_fragment_len, length)
            upper = total - length + 1
            if self.params.prioritize_ends:
                upper += self.params.min_fragment_len
            index = min(np.random.randint(upper), total - length)
            assert index >= 0  # should be impossible, just double checking

            # sequence['is_first'] = np.zeros(len(sequence['action']), np.bool)
            # sequence['is_first'][0] = True

            # TODO: this won't work if we use a subclass of SequenceData.
            # Should make a func like attr.map_structure or something
            data = map_structure(lambda x: x[index : index + length], attr.asdict(episode))
            assert self.params.min_fragment_len <= len(data["done"]) <= self.params.max_fragment_len
            return SequenceData(**data)

    def __iter__(self) -> Iterator:
        while True:
            yield self.get_trajectory()



from typing import ClassVar
from typing import Final
from typing import List
from typing import Optional
from typing import Type

import attr
from godot_parser import Node as GDNode

from avalon.datagen.world_creation.entities.constants import LARGEST_ANIMAL_SIZE
from avalon.datagen.world_creation.entities.item import InstancedDynamicItem
from avalon.datagen.world_creation.types import GodotScene


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Animal(InstancedDynamicItem):
    is_grounded: bool = False
    is_flying: bool = False
    is_able_to_climb: bool = False

    detection_radius_override: Optional[float] = None

    def get_offset(self):
        return LARGEST_ANIMAL_SIZE / 2

    def get_node(self, scene: GodotScene) -> GDNode:
        node = super().get_node(scene)
        if self.detection_radius_override is not None:
            node.properties["player_detection_radius"] = self.detection_radius_override
        return node


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Prey(Animal):
    RESOURCES: ClassVar = "res://entities/animals/prey"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Predator(Animal):
    RESOURCES: ClassVar = "res://entities/animals/predators"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Frog(Prey):
    resource_file: str = f"{Prey.RESOURCES}/frog.tscn"
    is_grounded: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Turtle(Prey):
    resource_file: str = f"{Prey.RESOURCES}/turtle.tscn"
    is_grounded: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Mouse(Prey):
    resource_file: str = f"{Prey.RESOURCES}/mouse.tscn"
    is_grounded: bool = True
    is_able_to_climb: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Rabbit(Prey):
    resource_file: str = f"{Prey.RESOURCES}/rabbit.tscn"
    is_grounded: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Pigeon(Prey):
    resource_file: str = f"{Prey.RESOURCES}/pigeon.tscn"
    is_flying: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Squirrel(Prey):
    resource_file: str = f"{Prey.RESOURCES}/squirrel.tscn"
    is_grounded: bool = True
    is_able_to_climb: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Crow(Prey):
    resource_file: str = f"{Prey.RESOURCES}/crow.tscn"
    is_flying: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Deer(Prey):
    resource_file: str = f"{Prey.RESOURCES}/deer.tscn"
    is_grounded: bool = True
    mass: float = 3.0


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Bee(Predator):
    resource_file: str = f"{Predator.RESOURCES}/bee.tscn"
    is_flying: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Snake(Predator):
    resource_file: str = f"{Predator.RESOURCES}/snake.tscn"
    is_grounded: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Hawk(Predator):
    resource_file: str = f"{Predator.RESOURCES}/hawk.tscn"
    is_flying: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Hippo(Predator):
    resource_file: str = f"{Predator.RESOURCES}/hippo.tscn"
    is_grounded: bool = True
    mass: float = 10.0


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Alligator(Predator):
    resource_file: str = f"{Predator.RESOURCES}/alligator.tscn"
    is_grounded: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Eagle(Predator):
    resource_file: str = f"{Predator.RESOURCES}/eagle.tscn"
    is_flying: bool = True


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Wolf(Predator):
    resource_file: str = f"{Predator.RESOURCES}/wolf.tscn"
    is_grounded: bool = True
    mass: float = 3.0


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Jaguar(Predator):
    resource_file: str = f"{Predator.RESOURCES}/jaguar.tscn"
    is_grounded: bool = True
    is_able_to_climb: bool = True
    mass: float = 3.0


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Bear(Predator):
    resource_file: str = f"{Predator.RESOURCES}/bear.tscn"
    is_grounded: bool = True
    is_able_to_climb: bool = True
    mass: float = 10.0


ALL_PREY_CLASSES: Final[List[Type[Prey]]] = [
    Frog,
    Turtle,
    Mouse,
    Rabbit,
    Pigeon,
    Squirrel,
    Crow,
    Deer,
]
ALL_PREDATOR_CLASSES: Final[List[Type[Predator]]] = [
    Bee,
    Snake,
    Hawk,
    Hippo,
    Alligator,
    Eagle,
    Wolf,
    Jaguar,
    Bear,
]



import attr
import numpy as np

from avalon.datagen.world_creation.entities.item import InstancedItem


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Scenery(InstancedItem):
    is_dynamic: bool = False
    scale: np.ndarray = np.array([1, 1, 1])
    entity_id: int = -1



from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple

import attr
import numpy as np
from scipy import stats

from avalon.common.utils import only
from avalon.datagen.errors import WorldTooSmall
from avalon.datagen.world_creation.configs.building import BuildingConfig
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import IndoorTaskConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import AGENT_HEIGHT
from avalon.datagen.world_creation.constants import FOOD_HOVER_DIST
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.constants import CANONICAL_FOOD_HEIGHT_ON_TREE
from avalon.datagen.world_creation.entities.entity import Entity
from avalon.datagen.world_creation.entities.tools.placeholder import Placeholder
from avalon.datagen.world_creation.entities.tools.tool import Tool
from avalon.datagen.world_creation.indoor.builders import DefaultHallwayBuilder
from avalon.datagen.world_creation.indoor.builders import DefaultStoryLinker
from avalon.datagen.world_creation.indoor.builders import HouseLikeRoomBuilder
from avalon.datagen.world_creation.indoor.builders import RectangleFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import WindowBuilder
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.components import Room
from avalon.datagen.world_creation.indoor.components import Story
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.task_generator import CANONICAL_BUILDING_LOCATION
from avalon.datagen.world_creation.indoor.task_generator import BuildingTaskGenerator
from avalon.datagen.world_creation.indoor.task_generator import IndoorTaskParams
from avalon.datagen.world_creation.indoor.task_generator import create_building_for_skill_scenario
from avalon.datagen.world_creation.indoor.task_generator import get_room_centroid_in_building_space
from avalon.datagen.world_creation.indoor.task_generator import make_indoor_task_world
from avalon.datagen.world_creation.indoor.task_generator import rectangle_dimensions_within_radius
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.types import HeightMode
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.height_path import HeightPath
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ClimbIndoorTaskConfig(IndoorTaskConfig):
    # Min/max site radii for a building with this task to be fit in
    min_site_radius: float = 7.0
    max_site_radius: float = 15.0
    # Min and max stories for the building to have at difficulty=0 and 1, respectively
    min_story_count: int = 2
    max_story_count: int = 5


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class ClimbOutdoorTaskConfig(TaskConfig):
    # only a particular path is made climbable, not the entire surface, which makes the task more challenging as this
    # path is made more narrow
    # the maximum number of "points" to include in the path (ie, how many segments it should have at difficulty 1.0)
    # note that this value cannot be larger than 2 due to implementation reasons (and making much more convoluted
    # paths starts to feel like a weird maze task, which isn't what we're going for)
    # a value of 0 means the climbable path will be a single straight line
    # easy corresponds to difficulty 0.0, hard to difficulty 1.0
    path_point_count_easy: int = 0
    path_point_count_hard: int = 2
    path_point_count_std_dev: float = 0.25
    # how wide to make the climbable path at difficulty 0.0 and 1.0 respectively
    path_width_easy: float = 10.0
    path_width_hard: float = 1.5
    path_width_std_dev: float = 0.1
    # bounds for over how much distance (in the Godot XZ plane) the climbing should take place
    min_climb_horizontal_dist: float = 1.0
    max_climb_horizontal_dist: float = 7.0
    # how high to make the cliff to be climbed. Note that it must be at least this tall in order to prevent just jumping
    # in order to accomplish the task
    cliff_height_easy: float = 3.2
    cliff_height_hard: float = 20.0
    # how steep to make the cliff for difficulty 0.0 and 1.0 respectively.
    # note that it cannot be less than this, or else you can walk up it, and it cannot be too steep, or else it becomes
    # difficult to have enough space for the multi-point paths on harder difficulties
    cliff_slope_easy: float = 1.43  # 55 degrees min
    cliff_slope_hard: float = 4.0  # 76 degrees max (mostly to keep climb_distance higher)


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class ClimbTaskConfig(TaskConfig):
    # likelihood that the indoor variant of the task will be used
    indoor_probability: float = 0.2
    # the configs for each variant of the task
    indoor_config: ClimbIndoorTaskConfig = ClimbIndoorTaskConfig()
    outdoor_config: ClimbOutdoorTaskConfig = ClimbOutdoorTaskConfig()


def generate_climb_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: ClimbTaskConfig = ClimbTaskConfig(),
) -> None:
    is_indoor = rand.uniform() < task_config.indoor_probability
    if is_indoor:
        building, entities, spawn_location, target_location = create_building_for_skill_scenario(
            rand,
            difficulty,
            ClimbTaskGenerator(task_config.indoor_config),
            position=CANONICAL_BUILDING_LOCATION,
            is_indoor_only=True,
        )
        world = make_indoor_task_world(
            building, entities, difficulty, spawn_location, target_location, rand, export_config
        )
    else:
        world, locations, difficulty = create_climb_obstacle(
            rand, difficulty, export_config, task_config=task_config.outdoor_config
        )
        world, locations = world.end_height_obstacles(
            locations, is_accessible_from_water=True, is_spawn_region_climbable=False
        )
        world = add_food_tree_for_simple_task(world, locations)
        world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)

    export_world(output_path, rand, world)


def create_climb_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    task_config: ClimbOutdoorTaskConfig = ClimbOutdoorTaskConfig(),
) -> Tuple[World, WorldLocations, float]:

    is_everywhere_climbable, difficulty = select_boolean_difficulty(difficulty, rand)
    # for the agent we may want to move height here because its difficulty may be more important than path_point_count?
    path_point_count = round(
        normal_distrib_range(
            task_config.path_point_count_easy - 0.49,
            task_config.path_point_count_hard + 0.49,
            task_config.path_point_count_std_dev,
            rand,
            difficulty,
        )
    )

    # you aren't technically going to have to move this far--the food gets updated to be close to the cliff edge
    # this mostly just gives a bit of space for the climb paths
    desired_goal_dist = difficulty_variation(
        task_config.max_climb_horizontal_dist, task_config.max_climb_horizontal_dist * 4.0, rand, difficulty
    )

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, 0.5), rand, difficulty, export_config, constraint
    )

    critical_distance, _warnings = world.get_critical_distance(locations, task_config.min_climb_horizontal_dist)
    if critical_distance is None:
        raise WorldTooSmall(
            AvalonTask.CLIMB, task_config.min_climb_horizontal_dist, locations.get_2d_spawn_goal_distance()
        )

    max_climb_distance = min([critical_distance, task_config.max_climb_horizontal_dist])

    height = normal_distrib_range(task_config.cliff_height_easy, task_config.cliff_height_hard, 1.0, rand, difficulty)
    slope = normal_distrib_range(task_config.cliff_slope_easy, task_config.cliff_slope_hard, 0.5, rand, difficulty)
    climb_distance = min(
        normal_distrib_range(
            task_config.min_climb_horizontal_dist, max_climb_distance, max_climb_distance / 2.0, rand, difficulty
        ),
        height / slope,
    )
    path_width = normal_distrib_range(
        task_config.path_width_easy, task_config.path_width_hard, task_config.path_width_std_dev, rand, difficulty
    )

    outside_items: Tuple[Tool, ...] = tuple([Placeholder()])
    inside_items: Tuple[Tool, ...] = tuple()
    if constraint and constraint.is_height_inverted:
        outside_items, inside_items = inside_items, outside_items

    ring_config = make_ring(
        rand,
        difficulty,
        world,
        locations,
        constraint=constraint,
        gap_distance=0.0,
        height=height,
        traversal_width=normal_distrib_range(
            task_config.path_width_easy, task_config.path_width_hard, 1.0, rand, difficulty
        ),
        inner_traversal_length=climb_distance,
        is_single_obstacle=True,
        is_inside_climbable=is_everywhere_climbable,
        is_outside_climbable=is_everywhere_climbable,
        inner_solution=HeightSolution(
            paths=tuple()
            if is_everywhere_climbable
            else (
                HeightPath(
                    is_path_restricted_to_land=True,
                    extra_point_count=path_point_count,
                    width=path_width,
                    is_path_climbable=True,
                    is_height_affected=False,
                    # sometimes your paths will be slightly simpler than you expected.
                    # in those cases, a simplicity warning will be logged
                    is_path_failure_allowed=True,
                ),
            ),
            # this will be replaced with the food below
            outside_items=outside_items,
            inside_items=inside_items,
            # so that the food ends up away from the edge a little bit
            outside_item_radius=0.5,
            inside_item_radius=0.5,
            solution_point_brink_distance=1.5,
        ),
        height_mode=HeightMode.MIDPOINT_RELATIVE,
    )
    world = world.add_height_obstacle(rand, ring_config, locations.island)

    new_locations, world = replace_placeholder_with_goal(locations, world)

    return world, new_locations, difficulty


def replace_placeholder_with_goal(locations: WorldLocations, world: World) -> Tuple[WorldLocations, World]:
    # reset the location of the food to be where the placeholder ended up
    # height will be reset below
    placeholder_position = only([x for x in world.items if isinstance(x, Placeholder)]).position
    new_locations = attr.evolve(locations, goal=placeholder_position + UP_VECTOR * CANONICAL_FOOD_HEIGHT_ON_TREE)
    return new_locations, attr.evolve(world, items=tuple([x for x in world.items if not isinstance(x, Placeholder)]))


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ClimbTaskGenerator(BuildingTaskGenerator):
    config: ClimbIndoorTaskConfig = ClimbIndoorTaskConfig()

    def get_site_radius(self, rand: np.random.Generator, difficulty: float) -> float:
        return difficulty_variation(self.config.min_site_radius, self.config.max_site_radius, rand, difficulty)

    def get_building_config(
        self,
        rand: np.random.Generator,
        difficulty: float,
        radius: float,
        allowed_auxiliary_tasks: Tuple[BuildingTask, ...] = tuple(),
        aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    ) -> BuildingConfig:
        width, length = rectangle_dimensions_within_radius(radius)
        story_count = self.config.min_story_count + round(
            normal_distrib_range(0, self.config.max_story_count - self.config.min_story_count, 0.1, rand, difficulty)
        )
        return BuildingConfig(
            width=width,
            length=length,
            story_count=story_count,
            footprint_builder=RectangleFootprintBuilder(),
            room_builder=HouseLikeRoomBuilder(max_rooms=1),
            story_linker=DefaultStoryLinker(allow_ladders=True, allow_ramps=False),
            hallway_builder=DefaultHallwayBuilder(),
            window_builder=WindowBuilder(),
            aesthetics=aesthetics,
        )

    def get_principal_obstacle_params(self, rand: np.random.Generator, difficulty: float, building: Building) -> Tuple:
        spawn_story = building.stories[0]
        spawn_room = only(spawn_story.rooms)
        target_story = building.stories[-1]
        target_room = only(target_story.rooms)
        return spawn_story, spawn_room, target_story, target_room

    def add_principal_obstacles(
        self, rand: np.random.Generator, building: Building, obstacle_params: Tuple[Story, Room, Story, Room]
    ) -> IndoorTaskParams:
        spawn_story, spawn_room, target_story, target_room = obstacle_params
        entities: List[Entity] = []
        spawn_location = get_room_centroid_in_building_space(
            building, spawn_story, spawn_room, at_height=AGENT_HEIGHT / 2
        )
        target_location = get_room_centroid_in_building_space(
            building, target_story, target_room, at_height=FOOD_HOVER_DIST
        )
        entrance_sites = ((spawn_story.num, spawn_room.id, tuple(Azimuth)),)
        return building, entities, spawn_location, target_location, entrance_sites



"""A script that exercises the basic Avalon functionality to ensure the install was successful."""
from avalon.agent.godot.godot_gym import AvalonEnv
from avalon.agent.godot.godot_gym import GodotEnvironmentParams
from avalon.agent.godot.godot_gym import TrainingProtocolChoice
from avalon.common.log_utils import configure_local_logger

if __name__ == "__main__":
    configure_local_logger()

    env_params = GodotEnvironmentParams(
        resolution=96,
        training_protocol=TrainingProtocolChoice.SINGLE_TASK_EAT,
        initial_difficulty=0,
    )
    env = AvalonEnv(env_params)
    env.reset()

    def random_env_step():
        action = env.action_space.sample()
        obs, reward, done, info = env.step(action)
        if done:
            env.reset()
        return obs

    observations = [random_env_step() for _ in range(50)]

    print("avalon seems to be working properly!")  # script



"""
To maintain experiment reproducibility, this script is used to generate reference buildings / worlds in the exact state
that they were during data collection / experiments as published in the Avalon paper. These are then used for regression
testing to ensure refactoring / cleaning up code doesn't intentionally break reproducibility. (See test_indoor.py)

Whenever we release a new version of the environment, new references can be generated and committed.
"""
import argparse
import hashlib
import json
import shutil
import subprocess
import sys
from enum import Enum
from pathlib import Path
from typing import Callable
from typing import List
from typing import Tuple

import attr
import numpy as np

from avalon.common.errors import SwitchError
from avalon.common.log_utils import logger
from avalon.contrib.s3_utils import SimpleS3Client
from avalon.contrib.utils import TEMP_DIR
from avalon.contrib.utils import is_git_repo_clean_on_notebook
from avalon.contrib.utils import temp_dir
from avalon.datagen.world_creation.tests.fixtures import ChecksumManifest
from avalon.datagen.world_creation.tests.fixtures import get_current_reference_manifest
from avalon.datagen.world_creation.tests.helpers import create_indoor_world
from avalon.datagen.world_creation.tests.helpers import create_world
from avalon.datagen.world_creation.tests.helpers import export_building
from avalon.datagen.world_creation.tests.helpers import get_reference_data_path
from avalon.datagen.world_creation.tests.helpers import make_tarball
from avalon.datagen.world_creation.tests.helpers import make_world_file_generic
from avalon.datagen.world_creation.tests.params import BUILDING_CATALOG
from avalon.datagen.world_creation.tests.params import CANONICAL_SEED
from avalon.datagen.world_creation.tests.params import INDOOR_WORLD_CATALOG
from avalon.datagen.world_creation.tests.params import OUTDOOR_WORLD_CATALOG
from avalon.datagen.world_creation.worlds.world import build_building


def generate_reference_buildings(export_path_base: Path) -> None:
    building_id = 0
    for config_id, building_config in BUILDING_CATALOG.items():
        logger.info(f"Generating reference building {config_id}")

        building_name = f"{config_id}__{CANONICAL_SEED}"
        export_path = export_path_base / f"{building_name}.tscn"

        rand = np.random.default_rng(CANONICAL_SEED)
        try:
            building = build_building(building_config, building_id, rand)
            export_building(building, building_name, export_path)
        except Exception as e:
            logger.warning(f"{config_id} errored: {e}; writing empty file")
            # For any building configs that don't pan out (incompatible or broken) we write an empty file
            export_path.write_text("")


def generate_reference_indoor_worlds(export_path_base: Path) -> None:
    for world_name, params in INDOOR_WORLD_CATALOG.items():
        logger.info(f"Generating reference indoor world {world_name}")
        output_path = export_path_base / world_name
        output_path.mkdir()
        create_indoor_world(*params, export_path=output_path)


def generate_reference_outdoor_worlds(export_path_base: Path) -> None:
    for world_name, params in OUTDOOR_WORLD_CATALOG.items():
        logger.info(f"Generating reference outdoor world {world_name}")
        output_path = export_path_base / world_name
        output_path.mkdir()
        create_world(*params, export_path=output_path)
        for file in output_path.iterdir():
            make_world_file_generic(file)


def clean_path(path: Path, glob_pattern: str) -> None:
    for entry in path.glob(glob_pattern):
        if entry.is_dir():
            shutil.rmtree(entry)
        else:
            entry.unlink()


def archive_and_clean(source_path: Path, glob_pattern: str, archive_name: str) -> None:
    make_tarball(source_path, glob_pattern, source_path / archive_name)
    clean_path(source_path, glob_pattern)


def get_path_checksum(path: Path) -> str:
    md5 = hashlib.md5()
    for file in sorted(path.iterdir()):
        md5.update(file.read_bytes())
    return md5.hexdigest()


def get_file_checksum(file_path: Path) -> str:
    md5 = hashlib.md5()
    md5.update(file_path.read_bytes())
    return md5.hexdigest()


def get_git_revision_hash() -> str:
    return subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii").strip()


def make_checksum_json(source_path: Path, snapshot_commit: str, json_output_path: Path) -> None:
    checksums: ChecksumManifest = dict(snapshot_commit=snapshot_commit, checksums={})
    for entry in sorted(source_path.iterdir()):
        if entry.is_dir():
            checksums["checksums"][entry.stem] = get_path_checksum(entry)
        else:
            checksums["checksums"][entry.stem] = get_file_checksum(entry)
    return json.dump(checksums, open(json_output_path, "w"), indent=2)


class Mode(Enum):
    REVIEW = "review"
    UPDATE = "update"


@attr.s(auto_attribs=True, eq=False, collect_by_mro=True)
class DataReference:
    name: str
    generate_data: Callable[[Path], None]
    glob_pattern: str


def make_reference_review(reference: DataReference, data_path: Path) -> Tuple[Path, Path]:
    review_path = data_path / "reference_review" / reference.name
    review_current_path = review_path / "current"
    review_new_path = review_path / "new"
    assert not review_path.exists(), f"Review path {review_path} already exists, remove it before generating a new one"
    review_path.mkdir(parents=True)
    review_new_path.mkdir()

    try:
        reference_manifest = get_current_reference_manifest(reference.name)
        current_reference_path = get_reference_data_path(reference.name, reference_manifest["snapshot_commit"])
    except FileNotFoundError:
        logger.warning(f"warning: No reference manifest found for {reference.name}, first time generating these?")
        current_reference_path = None
    if current_reference_path is not None:
        shutil.copytree(current_reference_path, review_current_path)
    reference.generate_data(review_new_path)
    manifest_file_name = f"{reference.name}_manifest.json"
    manifest_path = review_new_path / manifest_file_name
    make_checksum_json(review_new_path, get_git_revision_hash(), manifest_path)
    return review_current_path, review_new_path


def update_reference(reference: DataReference, commit_hash: str, data_path: Path, upload_to_s3: bool = True) -> None:
    with temp_dir(TEMP_DIR) as export_path_base:
        reference.generate_data(export_path_base)

        manifest_file_name = f"{reference.name}_manifest.json"
        manifest_path = export_path_base / manifest_file_name
        make_checksum_json(export_path_base, commit_hash, manifest_path)
        (data_path / manifest_file_name).unlink(missing_ok=True)
        shutil.move(manifest_path, data_path)

        if upload_to_s3:
            s3_client = SimpleS3Client()
            tarball_name = f"{reference.name}_reference.tar.gz"
            tarball_path = export_path_base / tarball_name
            make_tarball(export_path_base, reference.glob_pattern, tarball_path)
            logger.info(f"Uploading {reference.name} to S3...")
            s3_client.upload_from_file(tarball_path, f"{commit_hash}/{tarball_name}")


def main(args: List[str]):
    parser = argparse.ArgumentParser()
    parser.add_argument("--mode", dest="mode", required=True, choices=[mode.value for mode in Mode])
    mode = Mode(parser.parse_args(args[1:]).mode)

    if mode == Mode.UPDATE and not is_git_repo_clean_on_notebook():
        logger.warning(
            "Warning: Asked to generate references, but git repo is not clean. Are you sure you want to do this?"
        )
        commit_hash = input("If so, enter the commit hash that these references should refer to:\n").strip()
    else:
        commit_hash = get_git_revision_hash()

    data_path = Path(__file__).parent / "data"
    references = [
        DataReference("buildings", generate_reference_buildings, "*.tscn"),
        DataReference("indoor_worlds", generate_reference_indoor_worlds, "indoor_world*"),
        DataReference("outdoor_worlds", generate_reference_outdoor_worlds, "outdoor_world*"),
    ]
    for reference in references:
        if mode == Mode.REVIEW:
            current_reference_path, new_reference_path = make_reference_review(reference, data_path)
            logger.info(
                f"Generated reference review data for {reference.name}\n"
                f"\tcurrent: {current_reference_path}\n"
                f"\tnew: {new_reference_path}"
            )
        elif mode == Mode.UPDATE:
            update_reference(reference, commit_hash, data_path)
            logger.info(f"Updated reference data and manifest for {reference.name}")
        else:
            raise SwitchError(f"Unknown mode: {mode}")


if __name__ == "__main__":
    main(sys.argv)



import torch
from torch import Tensor
from torch import nn


def test_forward_inference() -> Tensor:
    """Strange. We're implicitly using this file to test the plugin, because it won't type check without the plugin..."""
    layer = nn.Linear(16, 16)
    x = torch.rand((16,))
    y = layer(x)
    return y
